import numpy as np
import keras as ks
# import keras_core.saving


@ks.saving.register_keras_serializable(package='kgcnn', name='LeakySoftplus')
class LeakySoftplus(ks.layers.Layer):
    r"""Leaky softplus activation function similar to :obj:`tf.nn.leaky_relu` but smooth. """

    def __init__(self, alpha: float = 0.05, trainable: bool = False, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            alpha (float, optional): Leak parameter alpha. Default is 0.05.
            trainable (bool, optional): Whether set alpha trainable. Default is False.
        """
        super(LeakySoftplus, self).__init__(**kwargs)
        self._alpha_config = float(alpha)
        self._alpha_trainable = bool(trainable)
        self.alpha = self.add_weight(
            shape=tuple(),
            initializer=ks.initializers.Constant(alpha),
            dtype=self.dtype,
            trainable=self._alpha_trainable
        )

    def call(self, inputs, *args, **kwargs):
        """Forward pass.

        Args:
            inputs (Tensor): Input tenor of arbitrary shape.

        Returns:
            Tensor: Leaky soft-plus activation of inputs.
        """
        x = inputs
        return ks.activations.softplus(x) * (1 - self.alpha) + self.alpha * x

    def get_config(self):
        """Get layer config."""
        config = super(LeakySoftplus, self).get_config()
        config.update({"alpha": self._alpha_config, "trainable": self._alpha_trainable})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='LeakyRelu')
class LeakyRelu(ks.layers.Layer):
    r"""Leaky RELU function. Equivalent to :obj:`tf.nn.leaky_relu(x,alpha)` ."""

    def __init__(self, alpha: float = 0.05, trainable: bool = False, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            alpha (float, optional): Leak parameter alpha. Default is 0.05.
            trainable (bool, optional): Whether set alpha trainable. Default is False.
        """
        super(LeakyRelu, self).__init__(**kwargs)
        self._alpha_config = float(alpha)
        self._alpha_trainable = bool(trainable)
        self.alpha = self.add_weight(
            shape=tuple(), dtype=self.dtype,
            initializer=ks.initializers.Constant(alpha),
            trainable=self._alpha_trainable
        )

    def call(self, inputs, *args, **kwargs):
        """Forward pass.

        Args:
            inputs (Tensor): Input tenor of arbitrary shape.

        Returns:
            Tensor: Leaky relu activation of inputs.
        """
        x = inputs
        return ks.activations.leaky_relu(x, alpha=self.alpha)
        # return tf.nn.relu(x) - tf.nn.relu(-x)*self.alpha

    def get_config(self):
        """Get layer config."""
        config = super(LeakyRelu, self).get_config()
        config.update({"alpha": self._alpha_config, "trainable": self._alpha_trainable})
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='Swish')
class Swish(ks.layers.Layer):
    r"""Swish activation function. Computes :math:`x \; \text{sig}(\beta x)`,
    with :math:`\text{sig}(x) = 1/(1+e^{-x})`."""

    def __init__(self, beta: float = 1.0, trainable: bool = False, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            beta (float, optional): Parameter beta in sigmoid. Default is 1.0.
            trainable (bool, optional): Whether set beta trainable. Default is False.
        """
        super(Swish, self).__init__(**kwargs)
        self._beta_config = float(beta)
        self._beta_trainable = bool(trainable)
        self.beta = self.add_weight(
            shape=tuple(), dtype=self.dtype,
            initializer=ks.initializers.Constant(beta),
            trainable=self._beta_trainable
        )

    def call(self, inputs, *args, **kwargs):
        """Forward pass.

        Args:
            inputs (Tensor): Input tenor of arbitrary shape.

        Returns:
            Tensor: Swish activation of inputs.
        """
        x = inputs
        return x * ks.activations.sigmoid(self.beta * x)

    def get_config(self):
        """Get layer config."""
        config = super(Swish, self).get_config()
        config.update({"beta": self._beta_config, "trainable": self._beta_trainable})
        return config



import logging


def merge_metrics(m1, m2):
    """Merge two metric lists or dicts of `ks.metrics` objects."""
    if m1 is None:
        return m2
    if m2 is None:
        return m1
    # Dict case with multiple named outputs.
    if isinstance(m1, dict) and isinstance(m2, dict):
        keys = set(list(m1.keys()) + list(m2.keys()))
        m = {key: [] for key in keys}
        for mu in [m1, m2]:
            for key, value in mu.items():
                if value is not None:
                    m[key] = m[key] + (list(value) if isinstance(value, (list, tuple)) else [value])
        return m
    # Lists for single model output.
    m1 = [m1] if not isinstance(m1, (list, tuple)) else m1
    m2 = [m2] if not isinstance(m2, (list, tuple)) else m2
    if all([not isinstance(x1, (list, tuple)) for x1 in m1] + [not isinstance(x2, (list, tuple)) for x2 in m2]):
        return m1 + m2
    # List for multiple model output with nested lists.
    if len(m1) == len(m2):
        m = [[]] * len(m1)
        for i in range(len(m)):
            for mu in [m1, m2]:
                if mu[i] is not None:
                    m[i] = m[i] + (list(mu[i]) if isinstance(mu[i], (list, tuple)) else [mu[i]])
        return m
    else:
        logging.error("For multiple model outputs require same length of metrics list to merge.")
    logging.error("Can not merge metrics '%s' and '%s'." % (m1, m2))
    return None


import warnings
from copy import deepcopy, copy
import numpy as np
from scipy.spatial import Voronoi, ConvexHull
from networkx import MultiDiGraph
from pyxtal import pyxtal
from pymatgen.core.structure import Structure
from pymatgen.core.periodic_table import Element
from pymatgen.optimization.neighbors import find_points_in_spheres
from typing import Union, Optional, Any


def get_symmetrized_graph(structure: Union[Structure, pyxtal]) -> MultiDiGraph:
    """Builds a unit graph without any edges, but with symmetry information as node attributes.

    Each node has a `asymmetric_mapping` attribute,
    which contains the id of the symmetry-equivalent atom in the asymmetric unit.
    Each node has a `symmop` attribute,
    which contains the affine matrix to generate the position (in fractional coordinates) of the atom,
    from its symmetry-equivalent atom position in the asymmetric unit.
    Each node has a `multiplicity` attribute,
    which contains the multiplicity of the atom (how many symmetry-equivalent atoms there are for this node).
    The resulting graph will have a `spacegroup` attribute, that specifies the spacegroup of the crystal.

    Args:
        structure (Union[Structure, pyxtal]): Crystal structure to convert to a graph.

    Raises:
        ValueError: If the argument is not a pymatgen Structure or pyxtal object.

    Returns:
        MultiDiGraph: Unit graph with symmetry information, but without any edges for the crystal.
    """
    graph = MultiDiGraph()
    if isinstance(structure, pyxtal):
        pyxtal_cell = structure
    elif isinstance(structure, Structure):
        try:
            pyxtal_cell = pyxtal()
            pyxtal_cell.from_seed(structure)
        except:
            # use trivial spacegroup (with spacegroup number == 1)
            # if spglib isn't able to calculate symmetries
            frac_coords = np.array([site.frac_coords for site in structure.sites])
            frac_coords = _to_unit_cell(frac_coords)
            for node_idx, site in enumerate(structure.sites):
                graph.add_node(node_idx, atomic_number=site.specie.number,
                               asymmetric_mapping=node_idx,
                               frac_coords=frac_coords[node_idx],
                               coords=site.coords,
                               symmop=np.eye(4),
                               multiplicity=1)
            setattr(graph, 'lattice_matrix', structure.lattice.matrix)
            setattr(graph, 'spacegroup', 1)
            return graph
    else:
        raise ValueError("This method takes either a pymatgen.core.structure.Structure or a pyxtal object.")

    atomic_numbers, frac_coords, asymmetric_mapping, symmops, multiplicities = [], [], [], [], []
    for site in pyxtal_cell.atom_sites:
        atomic_numbers += (site.multiplicity * [Element(site.specie).Z])
        asymmetric_mapping += (site.multiplicity * [len(asymmetric_mapping)])
        frac_coords.append(site.coords)
        symmops += [symmop.affine_matrix for symmop in site.wp.ops]
        multiplicities += (site.multiplicity * [site.multiplicity])
    frac_coords = _to_unit_cell(np.vstack(frac_coords))
    lattice = pyxtal_cell.lattice.matrix
    coords = frac_coords @ lattice
    for node_idx in range(len(atomic_numbers)):
        graph.add_node(node_idx, atomic_number=atomic_numbers[node_idx],
                       asymmetric_mapping=asymmetric_mapping[node_idx],
                       frac_coords=frac_coords[node_idx],
                       coords=coords[node_idx],
                       symmop=symmops[node_idx],
                       multiplicity=multiplicities[node_idx])
    setattr(graph, 'lattice_matrix', lattice)
    setattr(graph, 'spacegroup', pyxtal_cell.group.number)

    return graph


def structure_to_empty_graph(structure: Union[Structure, pyxtal], symmetrize: bool = False) -> MultiDiGraph:
    """Builds an unit graph without any edges.

    Args:
        structure (Union[Structure, pyxtal]): Crystal structure to convert to a graph.
            symmetrize (bool, optional): Whether to include symmetry information attributes
            (`asymmetric_mapping`, `symmop`, `multiplicity` attributes) in nodes and graph
            (`spacegroup` atribute).
            Defaults to False.
        symmetrize (bool): Whether to get symmetrized graph.

    Raises:
        ValueError: If the argument is not a pymatgen Structure or pyxtal object.

    Returns:
        MultiDiGraph: Unit graph without any edges for the crystal.
    """
    if symmetrize:
        return get_symmetrized_graph(structure)
    else:
        if isinstance(structure, pyxtal):
            structure = structure.to_pymatgen()
        graph = MultiDiGraph()
        frac_coords = np.array([site.frac_coords for site in structure.sites])
        frac_coords = _to_unit_cell(frac_coords)
        for node_idx, site in enumerate(structure.sites):
            graph.add_node(node_idx, atomic_number=site.specie.number,
                           frac_coords=frac_coords[node_idx],
                           coords=site.coords,
                           **site.properties)
        setattr(graph, 'lattice_matrix', structure.lattice.matrix)
        return graph


def add_knn_bonds(graph: MultiDiGraph, k: int = 12, max_radius: float = 10.,
                  tolerance: Optional[float] = None, inplace: bool = False) -> MultiDiGraph:
    """Adds kNN-based edges to a unit cell graph.

    Args:
        graph (MultiDiGraph): The unit cell graph to add kNN-based edges to.
        k (int, optional): How many neighbors to add for each node. Defaults to 12.
        max_radius (float, optional): This parameter has no effect on the outcome of the graph.
            It may only on the runtime.
            The algorithm starts the kNN search in the environment the radius of max_radius.
            If the kth neighbor is not within this radius the algorithm is called again with twice the initial radius.
            Defaults to 10.
        tolerance (Optional[float], optional): If tolerance is not None,
            edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
            Defaults to None.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: Graph with added edges.
    """
    lattice = _get_attr_from_graph(graph, "lattice_matrix", make_copy=True)
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    coords = frac_coords @ lattice
    if max_radius is None:
        max_radius = _estimate_nn_radius_from_density(k, coords, lattice, 0.1)
    # return coords, lattice
    index1, index2, offset_vectors, distances = find_points_in_spheres(
        coords,
        coords,
        r=max_radius,
        pbc=np.array([True] * 3, dtype=int),
        lattice=lattice,
        tol=1e-8
    )
    offset_vectors = offset_vectors.astype('i2')
    # Remove self_loops:
    no_self_loops = np.argwhere(~np.isclose(distances, 0)).reshape(-1)
    index1 = index1[no_self_loops]
    index2 = index2[no_self_loops]
    offset_vectors = offset_vectors[no_self_loops]
    distances = distances[no_self_loops]

    new_graph = graph if inplace else deepcopy(graph)

    for node_idx in range(new_graph.number_of_nodes()):
        idxs = np.argwhere(index1 == node_idx)[:, 0]
        sorted_idxs = idxs[np.argsort(distances[idxs])]
        if len(sorted_idxs) < k:
            return add_knn_bonds(new_graph, k=k, max_radius=max_radius * 2, inplace=True)

        if tolerance is not None:
            cutoff = distances[sorted_idxs[k - 1]] + tolerance
            edge_idxs = idxs[np.argwhere(distances[idxs] <= cutoff)][:, 0]
        else:
            edge_idxs = sorted_idxs[:k]
        # If the max_radius doesn't capture k neighbors, try again with double the max_radius
        for edge_idx in edge_idxs:
            new_graph.add_edge(
                index2[edge_idx], index1[edge_idx],
                cell_translation=offset_vectors[edge_idx], distance=distances[edge_idx])

    return new_graph


def add_radius_bonds(graph: MultiDiGraph, radius: float = 5., inplace: bool = False) -> MultiDiGraph:
    """Adds radius-based edges to a unit cell graph.

    Args:
        graph (MultiDiGraph): The unit cell graph to add radius-based edges to.
        radius (float, optional): Cutoff radius for each atom in Angstrom units. Defaults to 5.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: Graph with added edges.
    """
    new_graph = graph if inplace else deepcopy(graph)

    lattice = _get_attr_from_graph(graph, "lattice_matrix", make_copy=True)
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    coords = frac_coords @ lattice
    index1, index2, offset_vectors, distances = find_points_in_spheres(
        coords, coords, r=radius, pbc=np.array([True] * 3, dtype=int), lattice=lattice, tol=1e-8)
    offset_vectors = offset_vectors.astype('i2')
    # Remove self_loops:
    no_self_loops = np.argwhere(~np.isclose(distances, 0)).reshape(-1)
    index1 = index1[no_self_loops]
    index2 = index2[no_self_loops]
    offset_vectors = offset_vectors[no_self_loops]
    distances = distances[no_self_loops]

    if len(index1) == 0:
        warnings.warn(
            'No edges added to the graph, consider increasing the radius and check your graph input instance.')

    for source, target, cell_translation, dist in zip(index2, index1, offset_vectors, distances):
        new_graph.add_edge(source, target, cell_translation=cell_translation, distance=dist)

    return new_graph


def add_voronoi_bonds(graph: MultiDiGraph, min_ridge_area: Optional[float] = None,
                      inplace: bool = False) -> MultiDiGraph:
    """Adds Voronoi-based edges to a unit cell graph.

    Args:
        graph (MultiDiGraph): The unit cell graph to add radius-based edges to.
        min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
            If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
            the atoms of the cells is excluded from the graph. Defaults to None.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: Graph with added edges.
    """
    new_graph = graph if inplace else deepcopy(graph)

    lattice = _get_attr_from_graph(graph, "lattice_matrix")
    frac_coords = np.array([data[1] for data in graph.nodes(data='frac_coords')])
    dim = lattice.shape[0]
    assert dim == 3
    size = np.array([1, 1, 1])
    expanded_frac_coords = _get_super_cell_grid_frac_coords(lattice, frac_coords, size)
    expanded_coords = expanded_frac_coords @ lattice
    flattened_expanded_coords = expanded_coords.reshape(-1, dim)

    voronoi = Voronoi(flattened_expanded_coords)
    ridge_points_unraveled = np.array(np.unravel_index(voronoi.ridge_points, expanded_coords.shape[:-1]))
    # shape: (num_ridges, 2 (source, target), 4 (3 cell_index + 1 atom_index))
    ridge_points_unraveled = np.moveaxis(ridge_points_unraveled, np.arange(dim), np.roll(np.arange(dim), 1))

    # Filter ridges that have source in the centered unit cell
    source_in_center_cell = np.argwhere(np.all(ridge_points_unraveled[:, 0, :dim] == 1, axis=-1))[:, 0]
    # Filter ridges that have target in the centered unit cell
    target_in_center_cell = np.argwhere(np.all(ridge_points_unraveled[:, 1, :dim] == 1, axis=-1))[:, 0]

    edge_info = np.vstack(
        [ridge_points_unraveled[source_in_center_cell][:, [1, 0]], ridge_points_unraveled[target_in_center_cell]])

    cell_translations = (edge_info[:, 0, :-1] - size).astype(float)
    edge_indices = edge_info[:, :, -1]

    distances = []
    for i in range(len(edge_indices)):
        d = np.linalg.norm(expanded_coords[tuple(edge_info[i][0])] - expanded_coords[tuple(edge_info[i][1])])
        distances.append(d)

    if min_ridge_area is not None:
        ridge_vertices = [voronoi.ridge_vertices[i] for i in
                          np.concatenate([source_in_center_cell, target_in_center_cell])]
        ridge_areas = [get_ridge_area(voronoi.vertices[idxs]) for idxs in ridge_vertices]
        for nodes, cell_translation, dist, ridge_area in zip(edge_indices, cell_translations, distances, ridge_areas):
            source, target = nodes[0], nodes[1]
            if ridge_area > min_ridge_area:
                new_graph.add_edge(source, target, cell_translation=cell_translation, distance=dist,
                                   voronoi_ridge_area=ridge_area)
    else:
        for nodes, cell_translation, dist in zip(edge_indices, cell_translations, distances):
            source, target = nodes[0], nodes[1]
            new_graph.add_edge(source, target, cell_translation=cell_translation, distance=dist)

    return new_graph


def remove_duplicate_edges(graph: MultiDiGraph, inplace=False) -> MultiDiGraph:
    """Removes duplicate edges with same offset.

    Args:
        graph (MultiDiGraph): The unit cell graph with edges to remove.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: The graph without duplicate edges.
    """
    new_graph = graph if inplace else deepcopy(graph)
    edge_counter = set()
    remove_edges = set()
    for e in new_graph.edges(data='cell_translation', keys=True):
        id_ = (e[0], e[1]) + tuple(e[3].astype(int))
        if id_ in edge_counter:
            remove_edges.add((e[0], e[1], e[2]))
        else:
            edge_counter.add(id_)
    for edge in remove_edges:
        new_graph.remove_edge(edge[0], edge[1], key=edge[2])
    return new_graph


def prune_knn_bonds(graph: MultiDiGraph, k: int = 12, tolerance: Optional[float] = None,
                    inplace: bool = False) -> MultiDiGraph:
    """Prunes edges of a graph to only the k with the smallest distance value.

    Args:
        graph (MultiDiGraph): The unit cell graph with edges to prune.
        k (int, optional): How many neighbors each node should maximally have. Defaults to 12.
        tolerance (Optional[float], optional): If tolerance is not None,
            edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
            Defaults to None.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: The graph with pruned edges.
    """
    new_graph = graph if inplace else deepcopy(graph)

    delete_edges = []
    for n in new_graph:
        edges = list(new_graph.in_edges(n, data='distance', keys=True))
        edges.sort(key=lambda x: x[3])
        if tolerance is not None:
            radius = edges[k][3] + tolerance
            delete_edges += [e[:3] for e in edges if e[3] > radius]
        else:
            delete_edges += [e[:3] for e in edges[k:]]
    new_graph.remove_edges_from(delete_edges)
    return new_graph


def prune_radius_bonds(graph: MultiDiGraph, radius: float = 4., inplace: bool = False) -> MultiDiGraph:
    """Prunes edges of a graph with larger distance than the specified radius.

    Args:
        graph (MultiDiGraph): The unit cell graph with edges to prune.
        radius (float, optional): Distance threshold. Edges with larger distance than this value are
            removed from the graph. Defaults to 4.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: The graph with pruned edges.
    """
    new_graph = graph if inplace else deepcopy(graph)

    delete_edges = []
    for e in new_graph.edges(data='distance', keys=True):
        if e[3] > radius:
            delete_edges.append(e[:3])
    new_graph.remove_edges_from(delete_edges)
    return new_graph


def prune_voronoi_bonds(graph: MultiDiGraph, min_ridge_area: Optional[float] = None,
                        inplace: bool = False) -> MultiDiGraph:
    """Prunes edges of a graph with a voronoi ridge are smaller then the specified min_ridge_area.

    Only works for graphs with edges that contain `voronoi_ridge_area` as edge attributes.
    Args:
        graph (MultiDiGraph): The unit cell graph with edges to prune.
        min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
                If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
                the atoms of the cells is excluded from the graph. Defaults to None.
        inplace (bool, optional): Whether to add the edges to the given graph or create a copy with added edges.
            Defaults to False.

    Returns:
        MultiDiGraph: The graph with pruned edges.
    """
    new_graph = graph if inplace else deepcopy(graph)

    if min_ridge_area is None:
        return new_graph

    delete_edges = []
    for e in new_graph.edges(data='voronoi_ridge_area', keys=True):
        if e[3] < min_ridge_area:
            delete_edges.append(e[:3])
    new_graph.remove_edges_from(delete_edges)
    return new_graph


def add_edge_information(graph: MultiDiGraph, inplace=False,
                         frac_offset=False, offset=True, distance=True) -> MultiDiGraph:
    """Adds edge information, such as offset ( `frac_offset`, `offset` ) and distances ( `distance` ) to edges.

    Args:
        graph (MultiDiGraph): Graph for which to add edge information.
        inplace (bool, optional): Whether to add the edge information to the given graph
            or create a copy with added edges.
            Defaults to False.
        frac_offset (bool, optional): Whether to add fractional offsets (`frac_offset` attribute) to edges.
            Defaults to False.
        offset (bool, optional): Whether to add offsets (`offset` attribute) to edges.
            Defaults to True.
        distance (bool, optional): Whether to add distances (`distance` attribute) to edges.
            Defaults to True.

    Returns:
        MultiDiGraph: The graph with added edge information.
    """
    new_graph = graph if inplace else deepcopy(graph)
    if graph.number_of_edges() == 0:
        return new_graph

    add_frac_offset = frac_offset
    add_offset = offset
    add_distance = distance

    frac_coords1 = []
    frac_coords2 = []
    cell_translations = []

    # Collect necessary coordinate information for calculations
    for e in new_graph.edges(data='cell_translation'):
        frac_coords1.append(new_graph.nodes[e[0]]['frac_coords'])
        cell_translations.append(e[2])
        frac_coords2.append(new_graph.nodes[e[1]]['frac_coords'])

    # Do calculations in vectorized form (instead of doing it inside the edge loop)
    frac_coords1 = np.array(frac_coords1)
    frac_coords2 = np.array(frac_coords2)
    cell_translations = np.array(cell_translations)
    frac_offset = frac_coords2 - (frac_coords1 + cell_translations)
    offset = frac_offset @ _get_attr_from_graph(new_graph, "lattice_matrix")
    if add_distance:
        distances = np.linalg.norm(offset, axis=-1)
    else:
        distances = None

    # Add calculated information to edge attributes
    for i, e in enumerate(new_graph.edges(data=True)):
        if add_frac_offset:
            e[2]['frac_offset'] = frac_offset[i]
        if add_offset:
            e[2]['offset'] = offset[i]
        if add_distance:
            e[2]['distance'] = distances[i]

    return new_graph


def to_non_periodic_unit_cell(graph: MultiDiGraph, add_reverse_edges: bool = True,
                              inplace: bool = False) -> MultiDiGraph:
    """Generates non-periodic graph representation from unit cell graph representation.

    Args:
        graph (MultiDiGraph): Unit cell graph to generate non-periodic graph for.
        add_reverse_edges (bool, optional): Whether to add incoming edges to atoms
            that lie outside the central unit cell.
            Defaults to True.
        inplace (bool, optional): Whether to add distances (`distance` attribute) to edges.
            Defaults to False.

    Returns:
        MultiDiGraph: Corresponding non-periodic graph for the given unit cell graph.
    """
    new_graph = graph if inplace else deepcopy(graph)
    new_nodes = dict()
    new_edges = []
    delete_edges = []
    node_counter = new_graph.number_of_nodes()
    for e in new_graph.edges(data=True, keys=True):
        cell_translation = e[3]['cell_translation']
        if np.any(cell_translation != 0):
            node_key = (e[0],) + tuple(cell_translation)
            if node_key not in new_nodes.keys():
                node_attrs = copy(new_graph.nodes[e[0]])
                node_attrs['frac_coords'] = new_graph.nodes[e[0]]['frac_coords'] + cell_translation
                node_attrs['coords'] = node_attrs['frac_coords'] @ _get_attr_from_graph(new_graph, "lattice_matrix")
                new_nodes[node_key] = (node_counter, node_attrs)
                node_counter += 1
            node_number, _ = new_nodes[node_key]
            edge_attrs1 = deepcopy(e[3])
            new_edges.append((node_number, e[1], edge_attrs1))
            if add_reverse_edges:
                edge_attrs2 = deepcopy(e[3])
                if 'frac_offset' in e[3].keys():
                    edge_attrs2['frac_offset'] = -e[3]['frac_offset']
                if 'offset' in e[3].keys():
                    edge_attrs2['offset'] = -e[3]['offset']
                new_edges.append((e[1], node_number, edge_attrs2))
            delete_edges.append((e[0], e[1], e[2]))
    new_graph.remove_edges_from(delete_edges)
    for node_number, node_attrs in new_nodes.values():
        new_graph.add_node(node_number, **node_attrs)
    for e in new_edges:
        new_graph.add_edge(e[0], e[1], **e[2])
    return new_graph


def to_supercell_graph(graph: MultiDiGraph, size) -> MultiDiGraph:
    """Generates super-cell graph representation from unit cell graph representation.

    Args:
        graph (MultiDiGraph): Unit cell graph to generate super cell graph for.
        size (list): How many cells the crystal will get expanded into each dimension.

    Returns:
        MultiDiGraph: Corresponding super cell graph for the given unit cell graph.
    """

    supercell_graph = MultiDiGraph()
    size_ = list(size) + [graph.number_of_nodes()]
    new_num_nodes = np.prod(size_)
    for node in range(new_num_nodes):
        idx = np.unravel_index(node, size_)
        cell_translation = idx[:3]
        node_num = idx[3]
        data = deepcopy(graph.nodes[node_num])
        data['frac_coords'] = data['frac_coords'] + np.array(cell_translation)
        data['coords'] = data['frac_coords'] @ _get_attr_from_graph(graph, "lattice_matrix")
        supercell_graph.add_node(node, **data)

    for edge in graph.edges(data=True):
        for cell_idx in range(np.prod(size)):
            cell_translation1 = np.unravel_index(cell_idx, size)
            cell_translation2 = (edge[2]['cell_translation'] + np.array(cell_translation1)).astype(int)
            if np.all(cell_translation2 >= 0) and np.all(cell_translation2 < size):
                new_source = np.ravel_multi_index(list(cell_translation2) + [edge[0]], size_)
                new_target = np.ravel_multi_index(list(cell_translation1) + [edge[1]], size_)
                data = deepcopy(edge[2])
                # del data['cell_translation']
                supercell_graph.add_edge(new_source, new_target, **data)

    setattr(supercell_graph, 'lattice_matrix', _get_attr_from_graph(graph, "lattice_matrix"))
    if hasattr(graph, 'spacegroup'):
        setattr(supercell_graph, 'spacegroup', graph.spacegroup)
    return supercell_graph


def to_asymmetric_unit_graph(graph: MultiDiGraph) -> MultiDiGraph:
    """Generates super cell graph representation from unit cell graph representation.

    Args:
        graph (MultiDiGraph): Unit cell graph to generate asymmetric unit graph for.

    Returns:
        MultiDiGraph: Corresponding asymmetric unit graph for the given unit cell graph.
    """

    asymmetric_mapping = np.array([node[1] for node in graph.nodes(data='asymmetric_mapping')])
    if None in asymmetric_mapping:
        raise ValueError(
            "".join([
                "Graph does not contain symmetry information. ",
                "Make sure to create the graph with `structure_to_empty_graph` ",
                "with the `symmetrize` argument set to `True` ."
            ])
        )
    asu_node_indice, inv_asymmetric_mapping = np.unique(asymmetric_mapping, return_inverse=True)

    asu_graph = MultiDiGraph()
    setattr(asu_graph, 'lattice_matrix', _get_attr_from_graph(graph, "lattice_matrix"))
    setattr(asu_graph, 'spacegroup', _get_attr_from_graph(graph, "spacegroup"))
    new_nodes_idx = {}

    # Add nodes of asymmetric unit to asu_graph
    for i, node_idx in enumerate(asu_node_indice):
        new_nodes_idx[node_idx] = i
        data = deepcopy(graph.nodes[node_idx])
        data['unit_cell_index'] = data['asymmetric_mapping']
        del data['asymmetric_mapping']
        del data['symmop']
        asu_graph.add_node(i, **data)

    if graph.number_of_edges() == 0:
        return asu_graph

    edges_to_keep = graph.in_edges(asu_node_indice, data=True)
    for e in edges_to_keep:
        source_data = graph.nodes[e[0]]
        new_source_idx = inv_asymmetric_mapping[e[0]]
        new_target_idx = new_nodes_idx[e[1]]
        data = deepcopy(e[2])
        data['symmop'] = source_data['symmop']
        asu_graph.add_edge(new_source_idx, new_target_idx, **data)

    return asu_graph


def _to_unit_cell(frac_coords):
    r"""Converts fractional coords to be within the :math:`[0,1)` interval.

    Args:
        frac_coords: Fractional coordinates to map into :math:`[0,1)` interval.

    Returns:
        Fractional coordinates within the [0,1) interval.
    """
    return frac_coords % 1. % 1.


def get_ridge_area(ridge_points):
    """Computes the ridge area given ridge points.

    Beware that this function, assumes that the ridge points are (roughly) within a flat subspace plane
    in the 3 dimensional space.
    It computes the area of the convex hull of the points in three dimensions and then divides it by two,
    since both sides of the flat convex hull are included.

    Args:
        ridge_points (np.ndarray): Ridge points to calculate area for.

    Returns:
        float: Ridge area for the given points.
    """
    while ridge_points.shape[0] <= 3:
        # Append copy of points to avoid QHull Error
        ridge_points = np.append(ridge_points, np.expand_dims(ridge_points[0], 0), 0)
    area = ConvexHull(ridge_points, qhull_options='QJ').area / 2
    return area


def pairwise_diff(coords1: np.ndarray, coords2: np.ndarray) -> np.ndarray:
    """Get the pairwise offset difference between two vector sets.

    Args:
        coords1 (np.ndarray): Coordinates of shape (..., n, 3)
        coords2 (np.ndarray): Coordinates of shape (..., m, 3)

    Returns:
        np.ndarray: Difference values of shape (..., n, m, 3)
    """
    # This can be solved more elegantly with normal broadcasting
    # TODO: Check if the same.
    def _reshape_at_axis(arr, axis, new_shape):
        # move reshape axis to last axis position, because np.reshape reshapes this first
        arr_tmp = np.moveaxis(arr, axis, -1)
        shape = arr_tmp.shape[:-1] + new_shape
        new_positions = np.arange(len(new_shape)) + axis
        old_positions = np.arange(len(new_shape)) + (len(arr.shape) - 1)
        # now call np.reshape and move axis to right position
        return np.moveaxis(arr_tmp.reshape(shape), old_positions, new_positions)

    # Difference calculated at last axis of both inputs
    assert coords1.shape[-1] == coords2.shape[-1]
    coords1_reshaped = coords1.reshape(-1, coords1.shape[-1])
    coords2_reshaped = coords2.reshape(-1, coords2.shape[-1])
    diffs = np.expand_dims(coords2_reshaped, 0) - np.expand_dims(coords1_reshaped, 1)
    return _reshape_at_axis(_reshape_at_axis(diffs, 1, coords2.shape[:-1]), 0, coords1.shape[:-1])


def _get_mesh(size: Union[int, list, tuple], dim: int) -> np.ndarray:
    """Utility function to create a numpy mesh grid with indices at last dimension.

    Args:
        size (int, list): Size of each dimension.
        dim (int): Dimension of the grid.

    Returns:
        np.ndarray: Mesh grid of form:
    """
    if isinstance(size, int):
        size = [size] * dim
    else:
        size = list(size)
    assert len(size) == dim

    mesh = np.array(np.meshgrid(*tuple([np.arange(i) for i in size])))
    mesh = np.moveaxis(mesh, [0, 1], [-1, 1])
    return mesh


def _get_cube(dim: int) -> np.ndarray:
    """Generate a cubic mesh.

    Args:
        dim (int): Dimension for cubic mesh.

    Returns:
        np.ndarray: Cubic mesh.
    """
    return _get_mesh(2, dim)


def _get_max_diameter(lattice: np.ndarray) -> Union[float, np.ndarray]:
    """Determine the max diameter of a lattice.

    Args:
        lattice (np.ndarray): Lattice matrix.

    Returns:
        np.ndarray: Max diameter of the lattice.
    """
    dim = lattice.shape[0]
    cube = _get_cube(dim)
    max_radius = np.max(np.linalg.norm((cube - 1 / 2) @ lattice, axis=1))
    return max_radius * 2


def _get_super_cell_grid_frac_coords(lattice: np.ndarray, frac_coords: np.ndarray, size: Union[int, list, np.ndarray]):
    """Get frac coordinates for positions in a grid of unit cells that is a cubic super-cell.

    ..code - block:: python

        import numpy as np
        from kgcnn.crystal.graph_builder import _get_super_cell_grid_frac_coords
        coordinates = _get_super_cell_grid_frac_coords(
            np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.5], [0.0, 0.5, 1.5]]),
            np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]),
            [3, 3, 3]
        )
        print(coordinates.shape)  # (7, 7, 7, 2, 3)

    Args:
        lattice (np.ndarray): Lattice matrix.
        frac_coords (np.ndarray): Fractional coordinates of atoms in unit cell.
        size (list): Size of the super-cell in each dimension.

    Returns:
        np.ndarray: List of fractional coordinates of atoms in the super-cell.
    """
    dim = lattice.shape[0]

    if isinstance(size, int):
        size = [size] * dim
    else:
        size = list(size)
    assert len(size) == dim

    doubled_size = np.array(size) * 2 + 1
    mesh = _get_mesh(doubled_size, dim)
    # frac_coords_expanded.shape == (1,1,1,num_atoms,3) (for dim == 3)
    # noinspection PyTypeChecker
    frac_coords_expanded = np.expand_dims(frac_coords, np.arange(dim).tolist())
    # mesh_expanded.shape == (double_size[0], double_size[1], double_size[2], 1, 3) (for dim == 3)
    mesh_expanded = np.expand_dims(mesh - size, -2)
    expanded_frac_coords = mesh_expanded + frac_coords_expanded

    return expanded_frac_coords


def _get_attr_from_graph(graph: MultiDiGraph, attr_name: str, make_copy: bool = False) -> Union[Any, np.ndarray]:
    """Utility function to obtain graph-level information of the underlying crystal.

    Args:
        graph (MultiDiGraph): Networkx graph object.
        attr_name (str): Name of the attribute.
        make_copy (bool): Copy crystal-graph attribute.

    Returns:
        np.ndarray: Crystal information.
    """
    if hasattr(graph, attr_name):
        if make_copy:
            out = deepcopy(getattr(graph, attr_name))
        else:
            out = getattr(graph, attr_name)
    else:
        raise AttributeError("Must attach attribute '%s' of crystal information to networkx graph." % attr_name)
    return out


def _estimate_nn_radius_from_density(k: int, coordinates: np.ndarray, lattice: np.ndarray,
                                     empirical_tol_factor: float = 0.0):
    """Rough estimate of the expected radius to find N nearest neighbours.

    Args:
        k (int): Number of neighbours.
        coordinates (np.ndarray): Coordinates array.
        lattice (np.ndarray): Lattice matrix.
        empirical_tol_factor (float): Tolerance factor for radius.

    Returns:
        float: estimated radius
    """
    volume_unit_cell = np.sum(np.abs(np.cross(lattice[0], lattice[1]) * lattice[2]))
    density_unit_cell = len(coordinates) / volume_unit_cell
    estimated_nn_volume = k / density_unit_cell  # + len(coordinates)/density_unit_cell
    estimated_nn_radius = abs(float(np.cbrt(estimated_nn_volume / np.pi * 3 / 4)))
    estimated_nn_radius = estimated_nn_radius * (1.0 + empirical_tol_factor)
    return estimated_nn_radius


def _get_geometric_properties_of_unit_cell(coordinates: np.ndarray, lattice: np.ndarray):
    """Diameter of a 3D unit cell and other properties.

    Args:
        coordinates (np.ndarray): Coordinates array.
        lattice (np.ndarray): Lattice matrix.

    Returns:
        tuple: (center_unit_cell, max_diameter_cell, volume_unit_cell, density_unit_cell)
    """
    # lattice_col = np.transpose(lattice)
    lattice_row = lattice
    center_unit_cell = np.sum(lattice_row, axis=0, keepdims=True) / 2  # (1, 3)
    max_radius_cell = np.amax(np.sqrt(np.sum(np.square(lattice_row - center_unit_cell), axis=-1)))
    max_diameter_cell = 2 * max_radius_cell
    volume_unit_cell = np.sum(np.abs(np.cross(lattice[0], lattice[1]) * lattice[2]))
    density_unit_cell = len(coordinates) / volume_unit_cell
    return center_unit_cell[0], max_diameter_cell, volume_unit_cell, density_unit_cell



from kgcnn.layers.modules import Embedding
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.gather import GatherNodesOutgoing
from keras.layers import Concatenate, Dense, Activation, Add, Dropout
from kgcnn.layers.gather import GatherState
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.pooling import PoolingNodes
from ._layers import DMPNNPPoolingEdgesDirected, GIN_D


def model_disjoint(
        inputs,
        use_node_embedding,
        use_edge_embedding,
        use_graph_embedding,
        use_graph_state=None,
        input_node_embedding=None,
        input_edge_embedding=None,
        input_graph_embedding=None,
        edge_initialize=None,
        edge_activation=None,
        edge_dense=None,
        depthDMPNN=None,
        dropoutDMPNN=None,
        pooling_args=None,
        gin_mlp=None,
        depthGIN=None,
        gin_args=None,
        output_embedding=None,
        node_pooling_kwargs=None,
        last_mlp=None,
        dropoutGIN=None,
        output_mlp=None
):
    n, ed, edi, batch_id_node, ed_pairs, count_nodes, graph_state = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)
    if use_graph_state:
        if use_graph_embedding:
            graph_state = Embedding(**input_graph_embedding)(graph_state)

    # Make first edge hidden h0 step 1
    h_n0 = GatherNodesOutgoing()([n, edi])
    h0 = Concatenate(axis=-1)([h_n0, ed])
    h0 = Dense(**edge_initialize)(h0)
    h0 = Activation(**edge_activation)(h0)  # relu equation 1

    # One Dense layer for all message steps this is not the case in DGIN they are independents!
    edge_dense_all = Dense(**edge_dense)  # see equation 3 comments

    # Model Loop steps 2 & 3
    h = h0
    for i in range(depthDMPNN):
        # equation 2
        m_vw = DMPNNPPoolingEdgesDirected()([n, h, edi, ed_pairs])  # ed_pairs for Directed Pooling!
        # equation 3
        h = edge_dense_all(m_vw)  # do one per layer ...
        # h = Dense(**edge_dense)(m_vw)
        h = Add()([h, h0])
        # remark : dropout before Activation in DGIN code
        h = Activation(**edge_activation)(h)
        if dropoutDMPNN is not None:
            h = Dropout(**dropoutDMPNN)(h)

    # equation 4 & 5
    m_v = AggregateLocalEdges(**pooling_args)([n, h, edi])
    m_v = Concatenate(axis=-1)([n, m_v])  #
    # equation 5b: hv = Dense(**node_dense)(mv) removed based on the paper

    # GIN_D part (this projection is normally not done in DGIN, but we need to get the correct "dim")
    n_units = gin_mlp["units"][-1] if isinstance(gin_mlp["units"], list) else int(gin_mlp["units"])
    h_v = Dense(n_units, use_bias=True, activation='linear')(m_v)
    h_v_0 = h_v

    list_embeddings = [h_v_0]  # empty in the paper
    for i in range(depthGIN):
        # not sure of the mv, mv ... here but why not ;-)
        h_v = GIN_D(**gin_args)(
            [h_v, edi, h_v_0])  # equation 6 & 7a  mv is new the new nodes values and we do pooling on ed via edi
        h_v = GraphMLP(**gin_mlp)([h_v, batch_id_node, count_nodes])  # equation 7b
        list_embeddings.append(h_v)

    # Output embedding choice look like it takes only the last h_v in the paper not all ???
    if output_embedding == 'graph':
        out = [
            PoolingNodes(**node_pooling_kwargs)([count_nodes, x, batch_id_node]) for x in list_embeddings
        ]  # will return tensor equation 8
        out = [MLP(**last_mlp)(x) for x in out]  # MLP apply per depthGIN
        if dropoutGIN is not None:
            out = [Dropout(**dropoutGIN)(x) for x in out]
        out = Add()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        if use_graph_state:
            graph_state_node = GatherState()([graph_state, batch_id_node])
            n = Concatenate()([n, graph_state_node])
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported graph embedding for mode `DGIN` .")

    return out



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from kgcnn.models.utils import update_model_kwargs
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from keras.backend import backend as backend_to_use
from kgcnn.layers.modules import Input
from ._model import model_disjoint

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023.11.15"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]

if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'AttentiveFP' is not supported." % backend_to_use())

# Implementation of AttentiveFP in `keras` from paper:
# Pushing the Boundaries of Molecular Representation for Drug Discovery with the Graph Attention Mechanism
# Zhaoping Xiong, Dingyan Wang, Xiaohong Liu, Feisheng Zhong, Xiaozhe Wan, Xutong Li, Zhaojun Li,
# Xiaomin Luo, Kaixian Chen, Hualiang Jiang*, and Mingyue Zheng*
# Cite this: J. Med. Chem. 2020, 63, 16, 8749–8760
# Publication Date:August 13, 2019
# https://doi.org/10.1021/acs.jmedchem.9b00959


model_default = {
    "name": "AttentiveFP",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None,), "name": "edge_number", "dtype": "int64"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "cast_disjoint_kwargs": {},
    "input_embedding": None,  # deprecated
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "input_edge_embedding": {"input_dim": 5, "output_dim": 64},
    "attention_args": {"units": 32},
    "depthmol": 2,
    "depthato": 2,
    "dropout": 0.1,
    "verbose": 10,
    "output_embedding": "graph",
    "output_scaling": None,
    "output_to_tensor": True,  # deprecated
    "output_tensor_type": "padded",
    "output_mlp": {"use_bias": [True, True, False], "units": [25, 10, 1],
                   "activation": ["relu", "relu", "sigmoid"]}
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               cast_disjoint_kwargs: dict = None,
               input_tensor_type: str = None,
               input_node_embedding: dict = None,
               input_edge_embedding: dict = None,
               input_embedding: dict = None,
               depthmol: int = None,
               depthato: int = None,
               dropout: float = None,
               attention_args: dict = None,
               name: str = None,
               verbose: int = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_tensor_type: str = None,
               output_scaling: dict = None,
               output_mlp: dict = None
               ):
    r"""Make `AttentiveFP <https://doi.org/10.1021/acs.jmedchem.9b00959>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.AttentiveFP.model_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, edges, edge_indices, ...]`
    with '...' indicating mask or id tensors following the template below:

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers if used.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        input_embedding (dict): Deprecated in favour of `input_node_embedding` etc.
        input_node_embedding (dict): Dictionary of arguments for nodes unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of arguments for edge unpacked in :obj:`Embedding` layers.
        depthato (int): Number of graph embedding units or depth of the network.
        depthmol (int): Number of graph embedding units or depth of the graph embedding.
        dropout (float): Dropout to use.
        attention_args (dict): Dictionary of layer arguments unpacked in :obj:`AttentiveHeadFP` layer. Units parameter
            is also used in GRU-update and :obj:`PoolingNodesAttentive`.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers.

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    di_inputs = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 1, 1],
        index_assignment=[None, None, 0]
    )

    n, ed, disjoint_indices, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = di_inputs

    # Wrapping disjoint model.
    out = model_disjoint(
        [n, ed, disjoint_indices, batch_id_node, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        use_edge_embedding=("int" in inputs[1]['dtype']) if input_edge_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        input_edge_embedding=input_edge_embedding,
        depthmol=depthmol,
        depthato=depthato,
        dropout=dropout,
        attention_args=attention_args,
        output_embedding=output_embedding,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import keras as ks
from keras.layers import Concatenate, Dense, Add, Activation, Dropout
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing, GatherState
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.literature.DMPNN._layers import DMPNNPPoolingEdgesDirected


def model_disjoint(inputs: list = None,
                   use_node_embedding: bool = None,
                   use_edge_embedding: bool = None,
                   use_graph_embedding: bool = None,
                   input_node_embedding: dict = None,
                   input_edge_embedding: dict = None,
                   input_graph_embedding: dict = None,
                   pooling_args: dict = None,
                   edge_initialize: dict = None,
                   edge_dense: dict = None,
                   edge_activation: dict = None,
                   node_dense: dict = None,
                   dropout: dict = None,
                   depth: int = None,
                   use_graph_state: bool = False,
                   output_embedding: str = None,
                   output_mlp: dict = None):
    n, ed, edi, batch_id_node, ed_pairs, count_nodes, graph_state = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)
    if use_graph_state:
        if use_graph_embedding:
            graph_state = Embedding(**input_graph_embedding)(graph_state)

    # Make first edge hidden h0
    h_n0 = GatherNodesOutgoing()([n, edi])
    h0 = Concatenate(axis=-1)([h_n0, ed])
    h0 = Dense(**edge_initialize)(h0)

    # One Dense layer for all message steps
    edge_dense_all = Dense(**edge_dense)  # Should be linear activation

    # Model Loop
    h = h0
    for i in range(depth):
        m_vw = DMPNNPPoolingEdgesDirected()([n, h, edi, ed_pairs])
        h = edge_dense_all(m_vw)
        h = Add()([h, h0])
        h = Activation(**edge_activation)(h)
        if dropout is not None:
            h = Dropout(**dropout)(h)

    mv = AggregateLocalEdges(**pooling_args)([n, h, edi])
    mv = Concatenate(axis=-1)([mv, n])
    hv = Dense(**node_dense)(mv)

    # Output embedding choice
    n = hv
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        if use_graph_state:
            out = ks.layers.Concatenate()([graph_state, out])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        if use_graph_state:
            graph_state_node = GatherState()([graph_state, n])
            n = Concatenate()([n, graph_state_node])
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported embedding mode for `DMPNN`.")

    return out



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint_weighted, model_disjoint_behler, model_disjoint_atom_wise
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-06"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'HDNNP2nd' is not supported." % backend_to_use())

# Implementation of HDNNP in `keras` from paper:
# Atom-centered symmetry functions for constructing high-dimensional neural network potentials
# by Jörg Behler (2011)
# https://aip.scitation.org/doi/abs/10.1063/1.3553717


model_default_weighted = {
    "name": "HDNNP2nd",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (None, 3), "name": "angle_indices_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
        {"shape": (), "name": "total_angles", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "cast_disjoint_kwargs": {},
    "has_charge_input": False,
    "w_acsf_ang_kwargs": {},
    "w_acsf_rad_kwargs": {},
    "normalize_kwargs": None,
    "const_normalize_kwargs": None,
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "predict_dipole": False,
    "use_output_mlp": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]},
    "output_tensor_type": "padded",
    "output_scaling": None
}


@update_model_kwargs(model_default_weighted, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model_weighted(inputs: list = None,
                        input_tensor_type: str = None,
                        cast_disjoint_kwargs: dict = None,
                        has_charge_input: bool = False,
                        node_pooling_args: dict = None,
                        name: str = None,
                        verbose: int = None,
                        w_acsf_ang_kwargs: dict = None,
                        w_acsf_rad_kwargs: dict = None,
                        normalize_kwargs: dict = None,
                        const_normalize_kwargs: dict = None,
                        mlp_kwargs: dict = None,
                        output_embedding: str = None,
                        use_output_mlp: bool = None,
                        output_to_tensor: bool = None,
                        predict_dipole: bool = None,
                        output_mlp: dict = None,
                        output_scaling: dict = None,
                        output_tensor_type: str = None
                        ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`__ graph network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_default_weighted` .
    Uses weighted `wACSF <https://arxiv.org/abs/1712.05861>`__ .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[node_number, coordinates, edge_indices, angle_indices, ...]`
    with '...' indicating mask or ID tensors following the template below.
    Requires node number for atom-wise neural networks.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        has_charge_input (bool): Whether the model needs total charge as input. Default is False.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        w_acsf_ang_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`wACSFAng` layer.
        w_acsf_rad_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`wACSFRad` layer.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphBatchNormalization` layer.
        const_normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFConstNormalization` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        predict_dipole (bool): Whether to predict additional dipole based on charges. Default is False.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0, 1, 2] + ([None] if has_charge_input else []),
        index_assignment=[None, None, 0, 0] + ([None] if has_charge_input else [])
    )

    if has_charge_input:
        n, x, disjoint_indices, ang_ind, tot_charge, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angle = dj
    else:
        n, x, disjoint_indices, ang_ind, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angle = dj
        tot_charge = None

    out = model_disjoint_weighted(
        [n, x, disjoint_indices, ang_ind, tot_charge, batch_id_node, count_nodes],
        node_pooling_args=node_pooling_args,
        w_acsf_ang_kwargs=w_acsf_ang_kwargs,
        w_acsf_rad_kwargs=w_acsf_rad_kwargs,
        normalize_kwargs=normalize_kwargs,
        const_normalize_kwargs=const_normalize_kwargs,
        mlp_kwargs=mlp_kwargs,
        output_embedding=output_embedding,
        use_output_mlp=use_output_mlp,
        output_mlp=output_mlp,
        predict_dipole=predict_dipole
    )

    if not isinstance(out, list):
        out = [out]

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        # We will only apply scale to first output, i.e. energy.
        out_scaled = out[0]
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out_scaled = scaler([out_scaled, n, batch_id_node])
        else:
            out_scaled = scaler(out_scaled)
        out[0] = out_scaled

    # Output embedding choice
    out = [template_cast_output(
        [out_to_cast, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    ) for out_to_cast in out]

    if len(out) == 1:
        out = out[0]

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model_weighted.__doc__ = make_model_weighted.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)

model_default_behler = {
    "name": "HDNNP2nd",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (None, 3), "name": "angle_indices_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
        {"shape": (), "name": "total_angles", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "cast_disjoint_kwargs": {},
    "has_charge_input": False,
    "g2_kwargs": {"eta": [0.0, 0.3], "rs": [0.0, 3.0], "rc": 10.0, "elements": [1, 6, 16]},
    "g4_kwargs": {"eta": [0.0, 0.3], "lamda": [-1.0, 1.0], "rc": 6.0,
                  "zeta": [1.0, 8.0], "elements": [1, 6, 16], "multiplicity": 2.0},
    "normalize_kwargs": {},
    "const_normalize_kwargs": None,
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "use_output_mlp": False,
    "predict_dipole": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]},
    "output_tensor_type": "padded",
    "output_scaling": None
}


@update_model_kwargs(model_default_behler, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model_behler(inputs: list = None,
                      input_tensor_type: str = None,
                      cast_disjoint_kwargs: dict = None,
                      has_charge_input: bool = None,
                      node_pooling_args: dict = None,
                      name: str = None,
                      verbose: int = None,
                      normalize_kwargs: dict = None,
                      const_normalize_kwargs: dict = None,
                      g2_kwargs: dict = None,
                      g4_kwargs: dict = None,
                      mlp_kwargs: dict = None,
                      output_embedding: str = None,
                      use_output_mlp: bool = None,
                      predict_dipole: bool = None,
                      output_to_tensor: bool = None,
                      output_mlp: dict = None,
                      output_scaling: dict = None,
                      output_tensor_type: str = None
                      ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`__ graph network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_default_behler` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[node_number, coordinates, edge_indices, angle_indices, ...]`
    with '...' indicating mask or ID tensors following the template below.
    Requires node number for atom-wise neural networks.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        has_charge_input (bool): Whether the model needs total charge as input. Default is False.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        g2_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFG2` layer.
        g4_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFG4` layer.
        normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphBatchNormalization` layer.
        const_normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFConstNormalization` layer.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        predict_dipole (bool): Whether to predict additional dipole based on charges. Default is False.
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0, 1, 2] + ([None] if has_charge_input else []),
        index_assignment=[None, None, 0, 0] + ([None] if has_charge_input else [])
    )

    if has_charge_input:
        n, x, disjoint_indices, ang_index, tot_charge, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angle = dj
    else:
        n, x, disjoint_indices, ang_index, batch_id_node, batch_id_edge, batch_id_angles, node_id, edge_id, angle_id, count_nodes, count_edges, count_angle = dj
        tot_charge = None

    out = model_disjoint_behler(
        [n, x, disjoint_indices, ang_index, tot_charge, batch_id_node, count_nodes],
        node_pooling_args=node_pooling_args,
        normalize_kwargs=normalize_kwargs,
        const_normalize_kwargs=const_normalize_kwargs,
        g2_kwargs=g2_kwargs,
        g4_kwargs=g4_kwargs,
        mlp_kwargs=mlp_kwargs,
        output_embedding=output_embedding,
        use_output_mlp=use_output_mlp,
        output_mlp=output_mlp,
        predict_dipole=predict_dipole
    )

    if not isinstance(out, list):
        out = [out]

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        # We will only apply scale to first output, i.e. energy.
        out_scaled = out[0]
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out_scaled = scaler([out_scaled, n, batch_id_node])
        else:
            out_scaled = scaler(out_scaled)
        out[0] = out_scaled

    # Output embedding choice
    out = [template_cast_output(
        [out_to_cast, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    ) for out_to_cast in out]

    if len(out) == 1:
        out = out[0]

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model_behler.__doc__ = make_model_behler.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)

model_default_atom_wise = {
    "name": "HDNNP2nd",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_representation", "dtype": "float32"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
    ],
    "input_tensor_type": "padded",
    "has_charge_input": False,
    "cast_disjoint_kwargs": {},
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "predict_dipole": False,
    "use_output_mlp": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]},
    "output_tensor_type": "padded",
    "output_scaling": None
}


@update_model_kwargs(model_default_atom_wise, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model_atom_wise(inputs: list = None,
                         input_tensor_type: str = None,
                         cast_disjoint_kwargs: dict = None,
                         has_charge_input: bool = None,
                         node_pooling_args: dict = None,
                         name: str = None,
                         verbose: int = None,
                         mlp_kwargs: dict = None,
                         output_embedding: str = None,
                         predict_dipole: bool = None,
                         use_output_mlp: bool = None,
                         output_to_tensor: bool = None,
                         output_mlp: dict = None,
                         output_scaling: dict = None,
                         output_tensor_type: str = None
                         ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`__ network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_default_atom_wise` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[node_number, node_representation, ...]`
    with '...' indicating mask or ID tensors following the template below.
    Requires node number for atom-wise neural networks.
    The representation are given directly to the model as they are expected to be pre-computed.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        has_charge_input (bool): Whether the model needs total charge as input. Default is False.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        predict_dipole (bool): Whether to predict additional dipole based on charges. Default is False.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0] + ([None] if has_charge_input else []),
        index_assignment=[None, None] + ([None] if has_charge_input else [])
    )

    if has_charge_input:
        n, x, tot_charge, batch_id_node, node_id, count_nodes = dj
    else:
        n, x, batch_id_node, node_id, count_nodes = dj
        tot_charge = None

    batch_id_edge, edge_id, count_edges = None, None, None

    out = model_disjoint_atom_wise(
        [n, x, tot_charge, batch_id_node, count_nodes],
        node_pooling_args=node_pooling_args,
        mlp_kwargs=mlp_kwargs,
        output_embedding=output_embedding,
        use_output_mlp=use_output_mlp,
        output_mlp=output_mlp,
        predict_dipole=predict_dipole
    )

    if not isinstance(out, list):
        out = [out]

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        # We will only apply scale to first output, i.e. energy.
        out_scaled = out[0]
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out_scaled = scaler([out_scaled, n, batch_id_node])
        else:
            out_scaled = scaler(out_scaled)
        out[0] = out_scaled

    # Output embedding choice
    out = [template_cast_output(
        [out_to_cast, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    ) for out_to_cast in out]

    if len(out) == 1:
        out = out[0]

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model_atom_wise.__doc__ = make_model_atom_wise.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)

# For default, the weighted ACSF are used, since they do should in principle work for all elements.
make_model = make_model_weighted
model_default = model_default_weighted



import keras as ks
from keras import ops


@ks.saving.register_keras_serializable(package='kgcnn', name='Adan')
class Adan(ks.optimizers.Optimizer):
    r"""Optimizer `Adan <https://arxiv.org/abs/2208.06677>`__ : Adaptive Nesterov Momentum Algorithm for
    Faster Optimizing Deep Models.

    'Adan develops a Nesterov momentum estimation method to estimate stable and accurate first
    and second momentums of gradient in adaptive gradient algorithms for acceleration'.

    Algorithm of Adan:

    Input: Initialization :math:`θ_0`, step size :math:`\eta`, average parameter
    :math:`(β_1, β_2, β_3) \in [0, 1]^3`, stable parameter :math:`\epsilon > 0`,
    weight decays :math:`\lambda_k > 0`, restart condition.

    Output: some average of :math:`\{\theta_k\}^K_{k=1}`.

    (set :math:`m_0 = g_0` and :math:`v_1 = g_1 - g_0`)


    while :math:`k < K` do:

    .. math::

        m_k &= (1 − \beta_1)m_{k−1} + \beta_1 g_k  \\\\
        v_k &= (1 − \beta_2)v_{k−1} + \beta_2(g_k − g_{k−1}) \\\\
        n_k = (1 − \beta_3)n_{k−1} + \beta_3[g_k + (1 − \beta_2)(g_k − g_{k−1})]^2 \\\\
        \eta_k = \eta / \sqrt{n_k + \epsilon}  \\\\
        θ_{k+1} = (1 + \lambda_k \eta)^{-1} [\theta_k − \eta_k \dot (m_k + (1 − \beta_2) v_k)] \\\\
        \text{if restart condition holds:} \\\\
        \text{   get stochastic gradient estimator } g_0 \text{at } \theta_{k+1} \\\\
        \text{   set } m_0 = g_0, \; v_0 = 0, \; n_0 = g_0^2, \; k = 1 \\\\
        \text{   update } \theta_k
    """

    # Reference pytorch implementations:
    # https://github.com/frgfm/Holocron/blob/main/holocron/optim/functional.py
    # https://github.com/sail-sg/Adan/blob/main/adan.py
    # https://github.com/lucidrains/Adan-pytorch

    def __init__(self,
                 learning_rate: float = 1e-3,
                 name: str = "Adan",
                 beta_1: float = 0.98,
                 beta_2: float = 0.92,
                 beta_3: float = 0.99,
                 eps: float = 1e-8,
                 weight_decay: float = 0.0,
                 amsgrad: bool = False,
                 clipnorm=None,
                 clipvalue=None,
                 global_clipnorm=None,
                 use_ema=False,
                 ema_momentum=0.99,
                 ema_overwrite_frequency=None,
                 **kwargs):
        """Initialize optimizer.

        Args:
            learning_rate (float): Learning rate. Default is 1e-3.
            name (str): Name of the optimizer. Defaults to 'Adan'.
            beta_1 (float): Beta 1 parameter. Default is 0.98.
            beta_2 (float): Beta 2 parameter. Default is 0.92.
            beta_3 (float): Beta 3 parameter. Default is 0.99.
            eps (float): Numerical epsilon for denominators. Default is 1e-8.
            weight_decay (float): Decoupled weight decay. Default is 0.0.
            amsgrad (bool): Use the maximum of all 2nd moment running averages. Default is False.
        """
        super(Adan, self).__init__(
            learning_rate=learning_rate,
            name=name,
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            **kwargs)

        if not 0.0 <= learning_rate:
            raise ValueError("Invalid learning rate: {}".format(learning_rate))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= beta_1 < 1.0:
            raise ValueError("Invalid beta_1 parameter: {}".format(beta_1))
        if not 0.0 <= beta_2 < 1.0:
            raise ValueError("Invalid beta_2 parameter: {}".format(beta_2))
        if not 0.0 <= beta_3 < 1.0:
            raise ValueError("Invalid beta_3 parameter: {}".format(beta_3))

        self._input_learning_rate = float(learning_rate)
        self._eps = float(eps)
        self._beta_1 = float(beta_1)
        self._beta_2 = float(beta_2)
        self._beta_3 = float(beta_3)
        self._input_weight_decay = weight_decay
        self._use_amsgrad = bool(amsgrad)

    def build(self, var_list):
        """Initialize optimizer variables.

        Args:
            var_list: list of model variables to build Adam variables on.
        """
        if self.built:
            return
        super().build(var_list)
        self._exp_avg = []
        self._exp_avg_sq = []
        self._exp_avg_diff = []
        self._pre_grad = []
        for var in var_list:
            self._exp_avg.append(
                self.add_variable_from_reference(
                    reference_variable=var, name="exp_avg"
                )
            )
            self._exp_avg_sq.append(
                self.add_variable_from_reference(
                    reference_variable=var, name="exp_avg_sqs"
                )
            )
            self._exp_avg_diff.append(
                self.add_variable_from_reference(
                    reference_variable=var, name="exp_avg_diff"
                )
            )
            self._pre_grad.append(
                self.add_variable_from_reference(
                    reference_variable=var, name="pre_grad"
                )
            )

        if self._use_amsgrad:
            self._max_exp_avg_sq = []
            for var in var_list:
                self._max_exp_avg_sq.append(
                    self.add_variable_from_reference(
                        reference_variable=var, name="max_exp_avg_sq"
                    )
                )

    def update_step(self, grad, var, learning_rate):
        """Update step given gradient and the associated model variable."""
        var_dtype = var.dtype
        lr_t = ops.cast(learning_rate, var_dtype)  # lr_t = self._decayed_lr(var_dtype) done by super
        grad = ops.cast(grad, var_dtype)
        local_step = ops.cast(self.iterations + 1, var_dtype)

        beta1 = ops.cast(self._beta_1, var_dtype)
        beta2 = ops.cast(self._beta_2, var_dtype)
        beta3 = ops.cast(self._beta_3, var_dtype)
        bias_correction1 = 1 - ops.power(beta1, local_step)
        bias_correction2 = 1 - ops.power(beta2, local_step)
        bias_correction3 = 1 - ops.power(beta3, local_step)
        eps = ops.cast(self._eps, var_dtype)

        exp_avg = self._exp_avg[self._get_variable_index(var)]
        exp_avg_sq = self._exp_avg_sq[self._get_variable_index(var)]
        exp_avg_diff = self._exp_avg_diff[self._get_variable_index(var)]
        pre_grad = self._pre_grad[self._get_variable_index(var)]

        diff = grad - pre_grad

        self.assign(exp_avg, beta1 * exp_avg + grad * (1 - beta1))
        self.assign(exp_avg_diff, exp_avg_diff * beta2 + diff * (1 - beta2))
        update = grad + beta2 * diff
        self.assign(exp_avg_sq, exp_avg_sq * beta3 + update * update * (1 - beta3))

        if self._use_amsgrad:
            max_exp_avg_sq = self._max_exp_avg_sq[self._get_variable_index(var)]
            # Maintains the maximum of all 2nd moment running avg. till now
            self.assign(max_exp_avg_sq, ops.maximum(max_exp_avg_sq, exp_avg_sq))
            # Use the max. for normalizing running avg. of gradient
            denominator = (ops.sqrt(max_exp_avg_sq) / ops.sqrt(bias_correction3)) + eps
        else:
            denominator = (ops.sqrt(exp_avg_sq) / ops.sqrt(bias_correction3)) + eps

        update = (exp_avg / bias_correction1 + beta2 * exp_avg_diff / bias_correction2) / denominator

        self.assign_add(var, update * (-lr_t))

        self.assign(pre_grad, grad)

    def get_config(self):
        """Get config dictionary."""
        config = super(Adan, self).get_config()
        config.update(
            {
                "amsgrad": bool(self._use_amsgrad),
                "learning_rate": self._input_learning_rate,
                "eps": self._eps,
                "beta_1": self._beta_1,
                "beta_2": self._beta_2,
                "beta_3": self._beta_3,
                "weight_decay": self._input_weight_decay
            }
        )
        return config



import keras as ks
import keras.callbacks


class LearningRateLoggingCallback(ks.callbacks.Callback):
    """Callback logging the learning rate."""

    def __init__(self, verbose: int = 1):
        """Initialize class.

        Args:
            verbose (int): Verbosity. Default is 1.
        """
        super(LearningRateLoggingCallback, self).__init__()
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs=None):
        """Read out the learning rate on epoch end.

        Args:
            epoch (int): Number of current epoch.
            logs (dict): Dictionary of the logs.

        Returns:
            None.
        """
        lr = self.model.optimizer.learning_rate
        logs = logs or {}
        logs['lr'] = float(ks.backend.convert_to_numpy(self.model.optimizer.learning_rate))
        if self.verbose > 0:
            print("\nEpoch %05d: Finished epoch with learning rate: %s.\n" % (epoch + 1, logs['lr']))

    def get_config(self):
        """Get config for this class."""
        config = {"verbose": self.verbose}
        return config

    @classmethod
    def from_config(cls, config):
        """Make class instance from config."""
        return cls(**config)



import keras as ks
from kgcnn.layers.modules import Input
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use
from kgcnn.layers.scale import get as get_scaler
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from ._model import model_disjoint, model_disjoint_crystal

# To be updated if model is changed in a significant way.
__model_version__ = "2023-10-04"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]

if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'PAiNN' is not supported." % backend_to_use())

# Implementation of PAiNN in `keras` from paper:
# Equivariant message passing for the prediction of tensorial properties and molecular spectra
# Kristof T. Schuett, Oliver T. Unke and Michael Gastegger
# https://arxiv.org/pdf/2102.03150.pdf

model_default = {
    "name": "PAiNN",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 128},
    "has_equivariant_input": False,
    "equiv_initialize_kwargs": {"dim": 3, "method": "zeros", "units": 128},
    "bessel_basis": {"num_radial": 20, "cutoff": 5.0, "envelope_exponent": 5},
    "pooling_args": {"pooling_method": "scatter_sum"},
    "conv_args": {"units": 128, "cutoff": None, "conv_pool": "scatter_sum"},
    "update_args": {"units": 128, "add_eps": False},
    "equiv_normalization": False, "node_normalization": False,
    "depth": 3,
    "verbose": 10,
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "output_tensor_type": "padded",
    "output_scaling": None,
    "output_mlp": {"use_bias": [True, True], "units": [128, 1], "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               has_equivariant_input: bool = None,
               equiv_initialize_kwargs: dict = None,
               bessel_basis: dict = None,
               depth: int = None,
               pooling_args: dict = None,
               conv_args: dict = None,
               update_args: dict = None,
               equiv_normalization: bool = None,
               node_normalization: bool = None,
               name: str = None,
               verbose: int = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_scaling: dict = None,
               output_tensor_type: str = None
               ):
    r"""Make `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.PAiNN.model_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, ...]`
    with '...' indicating mask or ID tensors following the template below.
    If equivariant input is used via `has_equivariant_input` then input is extended to
    :obj:`[equiv, nodes, coordinates, edge_indices, ...]`

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        equiv_initialize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`EquivariantInitialize` layer.
        bessel_basis (dict): Dictionary of layer arguments unpacked in final :obj:`BesselBasisLayer` layer.
        depth (int): Number of graph embedding units or depth of the network.
        has_equivariant_input (bool): Whether the first equivariant node embedding is passed to the model.
        pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layer.
        conv_args (dict): Dictionary of layer arguments unpacked in :obj:`PAiNNconv` layer.
        update_args (dict): Dictionary of layer arguments unpacked in :obj:`PAiNNUpdate` layer.
        equiv_normalization (bool): Whether to apply :obj:`GraphLayerNormalization` to equivariant tensor update.
        node_normalization (bool): Whether to apply :obj:`GraphBatchNormalization` to node tensor update.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    disjoint_inputs = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=([0] if has_equivariant_input else []) + [0, 0, 1],
        index_assignment=([None] if has_equivariant_input else []) + [None, None, 0 + int(has_equivariant_input)]
    )

    if not has_equivariant_input:
        z, x, edi, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = disjoint_inputs
        v = None
    else:
        v, z, x, edi, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = disjoint_inputs

    # Wrapping disjoint model.
    out = model_disjoint(
        [z, x, edi, batch_id_node, batch_id_edge, count_nodes, count_edges, v],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        equiv_initialize_kwargs=equiv_initialize_kwargs,
        bessel_basis=bessel_basis, depth=depth, pooling_args=pooling_args, conv_args=conv_args,
        update_args=update_args, equiv_normalization=equiv_normalization, node_normalization=node_normalization,
        output_embedding=output_embedding, output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, z, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)


model_crystal_default = {
    "name": "PAiNN",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {'shape': (None, 3), 'name': "edge_image", 'dtype': 'int64'},
        {'shape': (3, 3), 'name': "graph_lattice", 'dtype': 'float32'},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 128},
    "has_equivariant_input": False,
    "equiv_initialize_kwargs": {"dim": 3, "method": "zeros"},
    "bessel_basis": {"num_radial": 20, "cutoff": 5.0, "envelope_exponent": 5},
    "pooling_args": {"pooling_method": "scatter_sum"},
    "conv_args": {"units": 128, "cutoff": None, "conv_pool": "scatter_sum"},
    "update_args": {"units": 128},
    "equiv_normalization": False,
    "node_normalization": False,
    "depth": 3,
    "verbose": 10,
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "output_scaling": None,
    "output_tensor_type": "padded",
    "output_mlp": {"use_bias": [True, True], "units": [128, 1], "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_crystal_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_crystal_model(inputs: list = None,
                       input_tensor_type: str = None,
                       input_embedding: dict = None,  # noqa
                       cast_disjoint_kwargs: dict = None,
                       has_equivariant_input: bool = None,
                       input_node_embedding: dict = None,
                       equiv_initialize_kwargs: dict = None,
                       bessel_basis: dict = None,
                       depth: int = None,
                       pooling_args: dict = None,
                       conv_args: dict = None,
                       update_args: dict = None,
                       equiv_normalization: bool = None,
                       node_normalization: bool = None,
                       name: str = None,
                       verbose: int = None,  # noqa
                       output_embedding: str = None,
                       output_to_tensor: bool = None,  # noqa
                       output_mlp: dict = None,
                       output_scaling: dict = None,
                       output_tensor_type: str = None
                       ):
    r"""Make `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.PAiNN.model_crystal_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edge_indices, image_translation, lattice, ...]`
    with '...' indicating mask or ID tensors following the template below.
    If equivariant input is used via `has_equivariant_input` then input is extended to
    :obj:`[equiv, nodes, coordinates, edge_indices, image_translation, lattice, ...]`

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        bessel_basis (dict): Dictionary of layer arguments unpacked in final :obj:`BesselBasisLayer` layer.
        equiv_initialize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`EquivariantInitialize` layer.
        depth (int): Number of graph embedding units or depth of the network.
        pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layer.
        has_equivariant_input (bool): Whether the first equivariant node embedding is passed to the model.
        conv_args (dict): Dictionary of layer arguments unpacked in :obj:`PAiNNconv` layer.
        update_args (dict): Dictionary of layer arguments unpacked in :obj:`PAiNNUpdate` layer.
        equiv_normalization (bool): Whether to apply :obj:`GraphLayerNormalization` to equivariant tensor update.
        node_normalization (bool): Whether to apply :obj:`GraphBatchNormalization` to node tensor update.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj_inputs = template_cast_list_input(
        model_inputs, input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=([0] if has_equivariant_input else []) + [
            0, 0, 1, 1, None],
        index_assignment=([0] if has_equivariant_input else []) + [
            None, None, 0 + int(has_equivariant_input), None, None]
    )

    if not has_equivariant_input:
        z, x, edi, img, lattice, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = dj_inputs
        v = None
    else:
        v, z, x, edi, img, lattice, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = dj_inputs

    # Wrapping disjoint model.
    out = model_disjoint_crystal(
        [z, x, edi, img, lattice, batch_id_node, batch_id_edge, count_nodes, count_edges, v],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        input_node_embedding=input_node_embedding, equiv_initialize_kwargs=equiv_initialize_kwargs,
        bessel_basis=bessel_basis, depth=depth, pooling_args=pooling_args, conv_args=conv_args,
        update_args=update_args, equiv_normalization=equiv_normalization, node_normalization=node_normalization,
        output_embedding=output_embedding, output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, z, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    model.__kgcnn_model_version__ = __model_version__
    return model


make_crystal_model.__doc__ = make_crystal_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



from ._make import make_model, model_default
from ._make import make_model_behler, model_default_behler
from ._make import make_model_weighted, model_default_weighted
from ._make import make_model_atom_wise, model_default_atom_wise

__all__ = [
    "make_model",
    "model_default",
    "make_model_behler",
    "model_default_behler",
    "make_model_weighted",
    "model_default_weighted",
    "make_model_atom_wise",
    "model_default_atom_wise"
]



import keras as ks
from keras import ops
from kgcnn.layers.aggr import AggregateLocalEdges
from keras.layers import Add, Multiply, Dense, Concatenate
from kgcnn.layers.geom import EuclideanNorm, ScalarProduct
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.modules import ExpandDims


class PAiNNconv(ks.layers.Layer):
    """Continuous filter convolution block of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ .

    Args:
        units (int): Units for Dense layer.
        conv_pool (str): Pooling method. Default is 'sum'.
        use_bias (bool): Use bias. Default is True.
        activation (str): Activation function. Default is 'kgcnn>shifted_softplus'.
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
    """

    def __init__(self, units,
                 conv_pool='scatter_sum',
                 use_bias=True,
                 activation='swish',
                 cutoff=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize Layer."""
        super(PAiNNconv, self).__init__(**kwargs)
        self.conv_pool = conv_pool
        self.units = units
        self.use_bias = use_bias
        self.cutoff = cutoff

        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        # Layer
        self.lay_dense1 = Dense(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_phi = Dense(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)
        self.lay_w = Dense(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)

        self.lay_split = SplitEmbedding(3, axis=-1)
        self.lay_sum = AggregateLocalEdges(pooling_method=conv_pool)
        self.lay_sum_v = AggregateLocalEdges(pooling_method=conv_pool)
        self.gather_n = GatherNodesOutgoing()
        self.gather_v = GatherNodesOutgoing()
        self.lay_mult = Multiply()
        if self.cutoff is not None:
            self.lay_mult_cutoff = Multiply()
        self.lay_exp_vv = ExpandDims(axis=-2)
        self.lay_exp_vw = ExpandDims(axis=-2)
        self.lay_exp_r = ExpandDims(axis=-1)
        self.lay_mult_vv = Multiply()
        self.lay_mult_vw = Multiply()

        self.lay_add = Add()

    def build(self, input_shape):
        """Build layer."""
        super(PAiNNconv, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: [nodes, equivariant, rbf, envelope, r_ij, edge_index]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - equivariant (Tensor): Equivariant node embedding of shape ([N], 3, F)
                - rdf (Tensor): Radial basis expansion pair-wise distance of shape ([M], #Basis)
                - envelope (Tensor): Distance envelope of shape ([N], 1)
                - r_ij (Tensor): Normalized pair-wise distance of shape ([M], 3)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            tuple: [ds, dv]

                - ds (Tensor) Updated node features of shape ([N], F)
                - dv (Tensor) Updated equivariant features of shape ([N], F, 3)
        """
        node, equivariant, rbf, envelope, r_ij, indexlist = inputs
        s = self.lay_dense1(node)
        s = self.lay_phi(s)
        s = self.gather_n([s, indexlist])
        w = self.lay_w(rbf)
        if self.cutoff is not None:
            w = self.lay_mult_cutoff([w, envelope])
        sw = self.lay_mult([s, w])
        sw1, sw2, sw3 = self.lay_split(sw, **kwargs)
        ds = self.lay_sum([node, sw1, indexlist])
        vj = self.gather_v([equivariant, indexlist])
        sw2 = self.lay_exp_vv(sw2)
        dv1 = self.lay_mult_vv([sw2, vj])
        sw3 = self.lay_exp_vw(sw3)
        r_ij = self.lay_exp_r(r_ij)
        dv2 = self.lay_mult_vw([sw3, r_ij])
        dv = self.lay_add([dv1, dv2])
        dv = self.lay_sum_v([node, dv, indexlist])
        return ds, dv

    def get_config(self):
        """Update layer config."""
        config = super(PAiNNconv, self).get_config()
        config.update({"conv_pool": self.conv_pool, "units": self.units, "cutoff": self.cutoff})
        config_dense = self.lay_dense1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            if x in config_dense:
                config.update({x: config_dense[x]})
        return config


class PAiNNUpdate(ks.layers.Layer):
    """Continuous filter convolution of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ .

    Args:
        units (int): Units for Dense layer.
        conv_pool (str): Pooling method. Default is 'sum'.
        use_bias (bool): Use bias. Default is True.
        activation (str): Activation function. Default is 'kgcnn>shifted_softplus'.
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
        add_eps: Whether to add eps in the norm.
    """

    def __init__(self, units,
                 use_bias=True,
                 activation='swish',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 add_eps: bool = False,
                 **kwargs):
        """Initialize Layer."""
        super(PAiNNUpdate, self).__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias
        self.add_eps = add_eps

        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}
        # Layer
        self.lay_dense1 = Dense(units=self.units, activation=activation, use_bias=self.use_bias, **kernel_args)
        self.lay_lin_u = Dense(self.units, activation='linear', use_bias=False, **kernel_args)
        self.lay_lin_v = Dense(self.units, activation='linear', use_bias=False, **kernel_args)
        self.lay_a = Dense(units=self.units * 3, activation='linear', use_bias=self.use_bias, **kernel_args)

        self.lay_scalar_prod = ScalarProduct(axis=1)
        self.lay_norm = EuclideanNorm(axis=1, add_eps=self.add_eps)
        self.lay_concat = Concatenate(axis=-1)
        self.lay_split = SplitEmbedding(3, axis=-1)

        self.lay_mult = Multiply()
        self.lay_exp_v = ExpandDims(axis=-2)
        self.lay_mult_vv = Multiply()
        self.lay_add = Add()

    def build(self, input_shape):
        """Build layer."""
        super(PAiNNUpdate, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: [nodes, equivariant]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - equivariant (Tensor): Equivariant node embedding of shape ([N], 3, F)

        Returns:
            tuple: [ds, dv]

                - ds (Tensor) Updated node features of shape ([N], F)
                - dv (Tensor) Updated equivariant features of shape ([N], 3, F)
        """
        node, equivariant = inputs
        v_v = self.lay_lin_v(equivariant, **kwargs)
        v_u = self.lay_lin_u(equivariant, **kwargs)
        v_prod = self.lay_scalar_prod([v_u, v_v], **kwargs)
        v_norm = self.lay_norm(v_v, **kwargs)
        a = self.lay_concat([node, v_norm], **kwargs)
        a = self.lay_dense1(a, **kwargs)
        a = self.lay_a(a, **kwargs)
        a_vv, a_sv, a_ss = self.lay_split(a, **kwargs)
        a_vv = self.lay_exp_v(a_vv, **kwargs)
        dv = self.lay_mult_vv([a_vv, v_u], **kwargs)
        ds = self.lay_mult([v_prod, a_sv], **kwargs)
        ds = self.lay_add([ds, a_ss], **kwargs)
        return ds, dv

    def get_config(self):
        """Update layer config."""
        config = super(PAiNNUpdate, self).get_config()
        config.update({"units": self.units, "add_eps": self.add_eps})
        config_dense = self.lay_dense1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            if x in config_dense.keys():
                config.update({x: config_dense[x]})
        return config


class EquivariantInitialize(ks.layers.Layer):
    """Equivariant initializer of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ .

    Args:
        dim (int): Dimension of equivariant features. Default is 3.
        method (str): How to initialize equivariant tensor. Default is "zeros".
    """

    def __init__(self, dim=3, units=128, method: str = "zeros", value: float = 1.0, stddev: float = 1.0,
                 seed: int = 42, **kwargs):
        """Initialize Layer."""
        super(EquivariantInitialize, self).__init__(**kwargs)
        self.dim = int(dim)
        self.units = units
        self.method = str(method)
        self.value = float(value)
        self.stddev = float(stddev)
        self.seed = seed

    def build(self, input_shape):
        """Build layer."""
        super(EquivariantInitialize, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass: Calculate edge update.

        Args:
            inputs: nodes

                - nodes (Tensor): Node embeddings of shape ([N], F) or ([N], ).

        Returns:
            Tensor: Equivariant tensor of shape ([N], dim, F) or ([N], dim, units).
        """
        if len(ops.shape(inputs)) < 2:
            inputs = ops.expand_dims(inputs, axis=-1)
            inputs = ops.repeat(inputs, self.units, axis=-1)

        if self.method == "zeros":
            out = ops.zeros_like(inputs, dtype=self.dtype)
            out = ops.expand_dims(out, axis=1)
            out = ops.repeat(out, self.dim, axis=1)
        elif self.method == "eps":
            out = ops.zeros_like(inputs, dtype=self.dtype) + ks.backend.epsilon()
            out = ops.expand_dims(out, axis=1)
            out = ops.repeat(out, self.dim, axis=1)
        elif self.method == "normal":
            out = ks.random.normal((self.dim, ops.shape(inputs)[-1]), seed=self.seed)
            out = ops.expand_dims(out, axis=0)
            out = ops.repeat(out, ops.shape(inputs)[0], axis=0)
        elif self.method == "ones":
            out = ops.ones_like(inputs, dtype=self.dtype)
            out = ops.expand_dims(out, axis=1)
            out = ops.repeat(out, self.dim, axis=1)
        elif self.method == "eye":
            out = ops.eye(self.dim, ops.shape(inputs)[1], dtype=self.dtype)
            out = ops.expand_dims(out, axis=0)
            out = ops.repeat(out, ops.shape(inputs)[0], axis=0)
        elif self.method == "const":
            out = ops.ones_like(inputs, dtype=self.dtype)*self.value
            out = ops.expand_dims(out, axis=1)
            out = ops.repeat(out, self.dim, axis=1)
        elif self.method == "node":
            out = ops.expand_dims(inputs, axis=1)
            out = ops.repeat(out, self.dim, axis=1)
            out = ops.cast(out, dtype=self.dtype)
        else:
            raise ValueError("Unknown initialization method %s" % self.method)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(EquivariantInitialize, self).get_config()
        config.update({"dim": self.dim, "method": self.method, "value": self.value, "stddev": self.stddev,
                       "units": self.units, "seed": self.seed})
        return config


class SplitEmbedding(ks.layers.Layer):
    """Split embeddings of `PAiNN <https://arxiv.org/pdf/2102.03150.pdf>`__ .

    Args:
        num_or_size_splits: Number or size of splits.
        axis (int): Axis to split.
        num (int): Number of output splits.
    """

    def __init__(self,
                 indices_or_sections,
                 axis=-1,
                 **kwargs):
        super(SplitEmbedding, self).__init__(**kwargs)
        self.indices_or_sections = indices_or_sections
        self.axis = axis

    def build(self, input_shape):
        super(SplitEmbedding, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass: Split embeddings across feature dimension e.g. `axis=-1` .

        Args:
            inputs (Tensor): Embeddings of shape ([N], F)

        Returns:
            list: List of tensor splits of shape ([N], F/num)
        """
        outs = ops.split(inputs, self.indices_or_sections, axis=self.axis)
        return outs

    def get_config(self):
        config = super(SplitEmbedding, self).get_config()
        config.update({"indices_or_sections": self.indices_or_sections, "axis": self.axis})
        return config



import kgcnn.backend as kgcnn_backend
from keras import KerasTensor
from kgcnn.backend import any_symbolic_tensors
from keras import Operation


class _ScatterMax(Operation):
    def call(self, indices, values, shape):
        return kgcnn_backend.scatter_reduce_max(indices, values, shape)

    def compute_output_spec(self, indices, values, shape):
        return KerasTensor(shape, dtype=values.dtype)


def scatter_reduce_max(indices, values, shape):
    r"""Scatter values at indices into new tensor of shape.

    Args:
        indices (Tensor): 1D Indices of shape `(M, )` .
        values (Tensor): Vales of shape `(M, ...)` .
        shape (tuple): Target shape.

    Returns:
        Tensor: Scattered values of `shape` .
    """
    if any_symbolic_tensors((indices, values, shape)):
        return _ScatterMax().symbolic_call(indices, values, shape)
    return kgcnn_backend.scatter_reduce_max(indices, values, shape)


class _ScatterMin(Operation):
    def call(self, indices, values, shape):
        return kgcnn_backend.scatter_reduce_min(indices, values, shape)

    def compute_output_spec(self, indices, values, shape):
        return KerasTensor(shape, dtype=values.dtype)


def scatter_reduce_min(indices, values, shape):
    r"""Scatter values at indices into new tensor of shape.

    Args:
        indices (Tensor): 1D Indices of shape `(M, )` .
        values (Tensor): Vales of shape `(M, ...)` .
        shape (tuple): Target shape.

    Returns:
        Tensor: Scattered values of `shape` .
    """
    if any_symbolic_tensors((indices, values, shape)):
        return _ScatterMin().symbolic_call(indices, values, shape)
    return kgcnn_backend.scatter_reduce_min(indices, values, shape)


class _ScatterMean(Operation):
    def call(self, indices, values, shape):
        return kgcnn_backend.scatter_reduce_mean(indices, values, shape)

    def compute_output_spec(self, indices, values, shape):
        return KerasTensor(shape, dtype=values.dtype)


def scatter_reduce_mean(indices, values, shape):
    r"""Scatter values at indices into new tensor of shape.

    Args:
        indices (Tensor): 1D Indices of shape `(M, )` .
        values (Tensor): Vales of shape `(M, ...)` .
        shape (tuple): Target shape.

    Returns:
        Tensor: Scattered values of `shape` .
    """
    if any_symbolic_tensors((indices, values, shape)):
        return _ScatterMean().symbolic_call(indices, values, shape)
    return kgcnn_backend.scatter_reduce_mean(indices, values, shape)


class _ScatterSum(Operation):
    def call(self, indices, values, shape):
        return kgcnn_backend.scatter_reduce_sum(indices, values, shape)

    def compute_output_spec(self, indices, values, shape):
        return KerasTensor(shape, dtype=values.dtype)


def scatter_reduce_sum(indices, values, shape):
    r"""Scatter values at indices into new tensor of shape.

    Args:
        indices (Tensor): 1D Indices of shape `(M, )` .
        values (Tensor): Vales of shape `(M, ...)` .
        shape (tuple): Target shape.

    Returns:
        Tensor: Scattered values of `shape` .
    """
    if any_symbolic_tensors((indices, values, shape)):
        return _ScatterSum().symbolic_call(indices, values, shape)
    return kgcnn_backend.scatter_reduce_sum(indices, values, shape)


class _ScatterSoftmax(Operation):

    def __init__(self, normalize: bool = False):
        super().__init__()
        self.normalize = normalize

    def call(self, indices, values, shape):
        return kgcnn_backend.scatter_reduce_softmax(indices, values, shape, normalize=self.normalize)

    def compute_output_spec(self, indices, values, shape):
        return KerasTensor(shape, dtype=values.dtype)


def scatter_reduce_softmax(indices, values, shape, normalize: bool = False):
    r"""Scatter values at indices to normalize values via softmax.

    Args:
        indices (Tensor): 1D Indices of shape `(M, )` .
        values (Tensor): Vales of shape `(M, ...)` .
        shape (tuple): Target shape of scattered tensor.

    Returns:
        Tensor: Values with softmax computed by grouping at indices.
    """
    if any_symbolic_tensors((indices, values, shape)):
        return _ScatterSoftmax(normalize=normalize).symbolic_call(indices, values, shape)
    return kgcnn_backend.scatter_reduce_softmax(indices, values, shape, normalize=normalize)


import keras as ks
from keras import ops
from keras.layers import Layer
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.aggr import AggregateLocalEdges
from keras.layers import Add, Concatenate


class rGIN(Layer):
    r"""Random Features Strengthen Graph Neural Networks <https://arxiv.org/abs/2002.03155>`__ .

    Computes graph convolution at step :math:`k` for node embeddings :math:`h_\nu` as:

    .. math::
        h_\nu^{(k)} = \phi^{(k)} ((1+\epsilon^{(k)}) h_\nu^{k-1} + \sum_{u\in N(\nu)}) h_u^{k-1}.

    with optional learnable :math:`\epsilon^{(k)}`

    .. note::

        The non-linear mapping :math:`\phi^{(k)}`, usually an :obj:`MLP`, is not included in this layer.
    """

    def __init__(self,
                 pooling_method='sum',
                 epsilon_learnable=False,
                 random_range=100,
                 **kwargs):
        """Initialize layer.

        Args:
            epsilon_learnable (bool): If epsilon is learnable or just constant zero. Default is False.
            pooling_method (str): Pooling method for summing edges. Default is 'segment_sum'.
        """
        super(rGIN, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.epsilon_learnable = epsilon_learnable
        self.random_range = random_range

        # Layers
        self.lay_gather = GatherNodesOutgoing()
        self.lay_pool = AggregateLocalEdges(pooling_method=self.pooling_method)
        self.lay_add = Add()
        self.lay_concat = Concatenate()

        # Epsilon with trainable as optional and default zeros initialized.
        self.eps_k = self.add_weight(
            name="epsilon_k",
            shape=(),
            trainable=self.epsilon_learnable,
            initializer="zeros",
            dtype=self.dtype
        )

    def build(self, input_shape):
        """Build layer."""
        super(rGIN, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [nodes, edge_index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - edge_index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Node embeddings of shape `([N], F)`
        """
        node, edge_index = inputs
        node_shape = ops.shape(node)
        random_values = ops.cast(
            ks.random.uniform([node_shape[0], 1], maxval=self.random_range, dtype="float32"),
        self.dtype) / self.random_range

        node = self.lay_concat([node, random_values])

        ed = self.lay_gather([node, edge_index], **kwargs)
        nu = self.lay_pool([node, ed, edge_index], **kwargs)  # Summing for each node connection
        no = (1 + self.eps_k) * node

        # Add the random features to the original features
        out = self.lay_add([node, nu], **kwargs)

        return out

    def get_config(self):
        """Update config."""
        config = super(rGIN, self).get_config()
        config.update({"pooling_method": self.pooling_method,
                       "epsilon_learnable": self.epsilon_learnable,
                       "random_range": self.random_range})
        return config



import numpy as np
from keras import ops
from kgcnn.utils.tests import TestCase
from kgcnn.layers.casting import CastBatchedIndicesToDisjoint, CastBatchedAttributesToDisjoint
from kgcnn.utils.tests import compare_static_shapes


class TestCastBatchedIndicesToDisjoint(TestCase):

    nodes = np.array([[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [1.0, 1.0]]])
    edges = np.array([[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]],
                      [[1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [-1.0, 1.0, 1.0]]])
    edge_indices = np.array([[[0, 0], [0, 1], [1, 0], [1, 1]],
                             [[0, 0], [0, 1], [1, 0], [1, 1]]], dtype="int64")
    node_mask = np.array([[True, False], [True, True]])
    edge_mask = np.array([[True, False, False, False], [True, True, True, False]])
    node_len = np.array([1, 2], dtype="int64")
    edge_len = np.array([1, 3], dtype="int64")

    def test_correctness(self):

        layer = CastBatchedIndicesToDisjoint(reverse_indices=True)
        layer_input = [self.nodes, ops.cast(self.edge_indices, dtype="int64"), self.node_len, self.edge_len]
        node_attr, edge_index, batch_node, batch_edge, node_id, edge_id, node_count, edge_count = layer(layer_input)
        self.assertAllClose(node_attr, [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0]])
        self.assertAllClose(edge_index, [[0, 1, 2, 1], [0, 1, 1, 2]])
        self.assertAllClose(batch_node, [0, 1, 1])
        self.assertAllClose(batch_edge, [0, 1, 1, 1])
        self.assertAllClose(node_id, [0, 0, 1])
        self.assertAllClose(edge_id, [0, 0, 1, 2])
        self.assertAllClose(node_count, [1, 2])
        self.assertAllClose(edge_count, [1, 3])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(None, 2), (2, None), (None, ), (None, ), (None, ), (None, ), (None, ), (None, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")

    def test_correctness_padding(self):

        layer = CastBatchedIndicesToDisjoint(padded_disjoint=True, reverse_indices=True)
        layer_input = [self.nodes, ops.cast(self.edge_indices, dtype="int64"), self.node_len, self.edge_len]
        node_attr, edge_index, batch_node, batch_edge, node_id, edge_id, node_count, edge_count = layer(layer_input)

        self.assertAllClose(node_attr, [[0.0, 0.0], [0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
        self.assertAllClose(edge_index, [[0, 1, 0, 0, 0, 3, 4, 3, 0], [0, 1, 0, 0, 0, 3, 3, 4, 0]])
        self.assertAllClose(batch_node, [0, 1, 0, 2, 2])
        self.assertAllClose(batch_edge, [0, 1, 0, 0, 0, 2, 2, 2, 0])
        self.assertAllClose(node_id, [0, 0, 0, 0, 1])
        self.assertAllClose(edge_id, [0, 0, 0, 0, 0, 0, 1, 2, 0])
        self.assertAllClose(node_count, [1, 1, 2])
        self.assertAllClose(edge_count, [4, 1, 3])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(5, 2), (2, 9), (5, ), (9, ), (5, ), (9, ), (3, ), (3, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")


class TestCastBatchedAttributesToDisjoint(TestCase):

    nodes = np.array([[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [1.0, 1.0]]])
    edges = np.array([[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]],
                      [[1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [-1.0, 1.0, 1.0]]])
    edge_indices = np.array([[[0, 0], [0, 1], [1, 0], [1, 1]],
                             [[0, 0], [0, 1], [1, 0], [1, 1]]], dtype="int64")
    node_mask = np.array([[True, False], [True, True]])
    edge_mask = np.array([[True, False, False, False], [True, True, True, False]])
    node_len = np.array([1, 2], dtype="int64")
    edge_len = np.array([1, 3], dtype="int64")

    def test_correctness(self):

        layer = CastBatchedAttributesToDisjoint(reverse_indices=True)
        layer_input = [self.nodes, self.node_len]
        node_attr, graph_id_node, node_id, node_len = layer(layer_input)
        self.assertAllClose(node_attr, [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0]])
        self.assertAllClose(graph_id_node, [0, 1, 1])
        self.assertAllClose(node_id, [0, 0, 1])
        self.assertAllClose(node_len, [1, 2])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(None, 2), (None, ), (None, ), (2, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")

        layer = CastBatchedAttributesToDisjoint(reverse_indices=True)
        layer_input = [self.edges, self.edge_len]
        edge_attr, graph_id_edge, edge_id, edge_len = layer(layer_input)
        self.assertAllClose(edge_attr, [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0]])
        self.assertAllClose(graph_id_edge, [0, 1, 1, 1])
        self.assertAllClose(edge_id, [0, 0, 1, 2])
        self.assertAllClose(edge_len, [1, 3])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(None, 3), (None, ), (None, ), (2, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")

    def test_correctness_padding(self):

        layer = CastBatchedAttributesToDisjoint(padded_disjoint=True, reverse_indices=True)
        layer_input = [self.nodes, self.node_len]
        node_attr, graph_id_node, node_id, node_len = layer(layer_input)
        self.assertAllClose(node_attr, [[0.0, 0.0], [0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
        self.assertAllClose(graph_id_node, [0, 1, 0, 2, 2])
        self.assertAllClose(node_id, [0, 0, 0, 0, 1])
        self.assertAllClose(node_len, [1, 1, 2])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(5, 2), (5, ), (5, ), (3, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")

        layer = CastBatchedAttributesToDisjoint(padded_disjoint=True)
        layer_input = [self.edges, self.edge_len]
        edge_attr, graph_id_edge, edge_id, edge_len = layer(layer_input)
        self.assertAllClose(edge_attr, [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0],
            [1.0, 1.0, 1.0], [1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [-1.0, 1.0, 1.0]])
        self.assertAllClose(graph_id_edge, [0, 1, 0, 0, 0, 2, 2, 2, 0])
        self.assertAllClose(edge_id, [0, 0, 0, 0, 0, 0, 1, 2, 0])
        self.assertAllClose(edge_len, [4, 1, 3])

        output_shape = layer.compute_output_shape([x.shape for x in layer_input])
        expected_output_shape = [(9, 3), (9, ), (9, ), (3, )]
        for f, e in zip(output_shape, expected_output_shape):
            self.assertTrue(compare_static_shapes(f, e), msg=f"Shape mismatch: {f} vs. {e}")


if __name__ == "__main__":

    TestCastBatchedIndicesToDisjoint().test_correctness()
    TestCastBatchedIndicesToDisjoint().test_correctness_padding()
    TestCastBatchedAttributesToDisjoint().test_correctness()
    TestCastBatchedAttributesToDisjoint().test_correctness_padding()
    print("Tests passed.")



import numpy as np
from keras import KerasTensor
from keras import Operation
from keras.backend import result_type
from kgcnn.backend import any_symbolic_tensors
import kgcnn.backend as kgcnn_backend
from kgcnn.ops.axis import broadcast_shapes


class _RepeatStaticLength(Operation):

    def __init__(self, total_repeat_length: int, axis=None):
        super().__init__()
        self.axis = axis
        self.total_repeat_length = total_repeat_length

    def call(self, x, repeats):
        return kgcnn_backend.repeat_static_length(
            x, repeats, total_repeat_length=self.total_repeat_length, axis=self.axis)

    def compute_output_spec(self, x, repeats):
        output_shape = list(x.shape)
        if self.axis is None:
            return KerasTensor([self.total_repeat_length], dtype=x.dtype)
        output_shape[self.axis] = self.total_repeat_length
        return KerasTensor(output_shape, dtype=x.dtype)


def repeat_static_length(x, repeats, total_repeat_length: int, axis=None):
    """Repeat each element of a tensor after themselves.

    Args:
        x: Input tensor.
        repeats: The number of repetitions for each element.
        total_repeat_length: length of all repeats.
        axis: The axis along which to repeat values. By default, use
            the flattened input array, and return a flat output array.

    Returns:
        Output tensor.
    """
    if any_symbolic_tensors((x, repeats)):
        return _RepeatStaticLength(axis=axis, total_repeat_length=total_repeat_length).symbolic_call(x, repeats)
    return kgcnn_backend.repeat_static_length(x, repeats, axis=axis, total_repeat_length=total_repeat_length)


class _DecomposeRaggedTensor(Operation):

    def __init__(self, batch_dtype: str = "int64"):
        super().__init__()
        self.batch_dtype = batch_dtype

    def call(self, x):
        kgcnn_backend.decompose_ragged_tensor(x)

    def compute_output_spec(self, x):
        x_shape = list(x.shape)
        output_shape = tuple(x_shape[1:])
        length_shape = (None, ) if x_shape[0] is None else (x_shape[0], )
        id_shape = output_shape[:1]
        return (KerasTensor(output_shape, dtype=x.dtype), KerasTensor(id_shape, dtype="int64"),
                KerasTensor(id_shape, dtype="int64"), KerasTensor(length_shape, dtype="int64"))


def decompose_ragged_tensor(x, batch_dtype="int64"):
    """Decompose ragged tensor.

    Args:
        x: Input tensor (ragged).
        batch_dtype (str): Data type for batch information. Default is 'int64'.

    Returns:
        Output tensors.
    """
    if any_symbolic_tensors((x,)):
        return _DecomposeRaggedTensor(batch_dtype=batch_dtype).symbolic_call(x)
    return kgcnn_backend.decompose_ragged_tensor(x)


def _reduce_shape(shape, axis=None, keepdims=False):
    shape = list(shape)
    if axis is None:
        if keepdims:
            return tuple([1 for _ in shape])
        else:
            return tuple([])

    if keepdims:
        for ax in axis:
            shape[ax] = 1
        return tuple(shape)
    else:
        for ax in sorted(axis, reverse=True):
            del shape[ax]
        return tuple(shape)


class _Norm(Operation):
    def __init__(self, ord='fro', axis=None, keepdims=False):
        super().__init__()
        self.ord = ord
        if isinstance(axis, int):
            axis = [axis]
        self.axis = axis
        self.keepdims = keepdims

    def call(self, x):
        return kgcnn_backend.norm(x, ord=self.ord, axis=self.axis, keepdims=self.keepdims)

    def compute_output_spec(self, x):
        return KerasTensor(
            _reduce_shape(x.shape, axis=self.axis, keepdims=self.keepdims),
            dtype=x.dtype,
        )


def norm(x, ord='fro', axis=None, keepdims=False):
    """Compute linalg norm.

    Args:
        x: Input tensor
        ord: Order of the norm.
        axis: dimensions over which to compute the vector or matrix norm.
        keepdims: If set to True, the reduced dimensions are retained in the result.

    Returns:
        output tensor.
    """
    if any_symbolic_tensors((x,)):
        return _Norm(ord=ord, axis=axis, keepdims=keepdims).symbolic_call(x)
    return kgcnn_backend.norm(x, ord=ord, axis=axis, keepdims=keepdims)


class _Cross(Operation):
    def __init__(self):
        super().__init__()
        self.axis = -1

    def call(self, x1, x2):
        return kgcnn_backend.cross(x1, x2)

    def compute_output_spec(self, x1, x2):
        x1_shape = list(x1.shape)
        x2_shape = list(x2.shape)

        x1_value_size = x1_shape[self.axis]
        x2_value_size = x2_shape[self.axis]
        del x1_shape[self.axis]
        del x2_shape[self.axis]
        output_shape = broadcast_shapes(x1_shape, x2_shape)

        if x1_value_size is not None and x1_value_size not in (2, 3):
            raise ValueError(
                "`x1`'s dim on `axis={axisa}` must be either 2 or 3, but "
                f"received: {x1_value_size}"
            )
        if x2_value_size is not None and x2_value_size not in (2, 3):
            raise ValueError(
                "`x2`'s dim on `axis={axisb}` must be either 2 or 3, but "
                f"received: {x2_value_size}"
            )

        if x1_value_size == 3 or x2_value_size == 3:
            value_size = [3]
        else:
            value_size = []

        output_shape = (
            output_shape[: self.axis] + value_size + output_shape[self.axis :]
        )
        dtype = result_type(x1.dtype, x2.dtype)
        return KerasTensor(output_shape, dtype=dtype)


def cross(x1, x2):
    """Returns the cross product of two (arrays of) vectors.

    Args:
        x1: Components of the first vector(s).
        x2: Components of the second vector(s).

    Returns:
        Vector cross product(s).
    """
    if any_symbolic_tensors((x1, x2)):
        return _Cross().symbolic_call(x1, x2)
    return kgcnn_backend.cross(x1, x2)
